"""Generic IDA Pro plugin for taskr framework."""

import idaapi

from . import DataProcessorCore, get_logger

logger = get_logger(__name__)


class DataProcessorPlugin(idaapi.plugin_t):
    """Generic IDA Pro plugin for data processing tasks."""

    flags = idaapi.PLUGIN_PROC
    comment = "Generic Data Processor via Shared Memory and Multiprocessing"
    help = "Press Alt-Shift-P to start processing"
    wanted_name = "TaskrProcessor"
    wanted_hotkey = "Alt-Shift-P"
    _core: DataProcessorCore | None = None

    def init(self):
        # This will be overridden by specific implementations
        logger.info("Generic TaskrProcessor plugin initialized")
        return idaapi.PLUGIN_KEEP

    def term(self):
        logger.info("Terminating plugin.")
        if self._core:
            self._core.terminate()
        logger.info("Plugin terminated.")

    def run(self, arg):
        # This will be overridden by specific implementations
        logger.warning("Generic plugin run() called - should be overridden")


def PLUGIN_ENTRY():
    return DataProcessorPlugin()


# XXX: This is a temporary hack to allow the plugin to be loaded
"""
from .ida_plugin import PLUGIN_ENTRY, DataProcessorPlugin

__all__ = ["DataProcessorPlugin", "PLUGIN_ENTRY"]
from ida_worker_manager import ConnectionContext, WorkerBase


class MyWorker(WorkerBase):
    def process(self, connection: ConnectionContext, **kwargs):
        # Your processing logic here
        connection.send_message("status", "processing")
        # Do work...
        connection.send_message("result", {"data": "results"})


# In your IDA plugin:
from ida_worker_manager import MessageHandler, WorkerLauncher


class MyHandler(MessageHandler):
    def on_worker_results(self, results):
        # Handle results
        pass


launcher = WorkerLauncher(MyHandler())
launcher.launch_worker("my_worker.py", {"arg1": "value1"})
"""



"""Worker-side components for the task worker manager."""

import asyncio
import concurrent.futures
import math
import multiprocessing
import multiprocessing.connection
import threading
import time
import typing
import uuid

from .helpers import get_logger
from .protocols import WorkerProtocol
from .utils import AsyncEventEmitter, IntervalSet
from .qt_compat import QT_ASYNCIO_AVAILABLE

logger = get_logger()

# Try to import QtAsyncio components if available
if QT_ASYNCIO_AVAILABLE:
    try:
        from .qtasyncio import set_event_loop_policy as qt_set_event_loop_policy
        QTASYNCIO_ENABLED = True
    except ImportError:
        QTASYNCIO_ENABLED = False
else:
    QTASYNCIO_ENABLED = False


class ConnectionContext:
    """
    Context manager for a multiprocessing.connection.Connection.
    Ensures the connection is closed on exit.
    """

    def __init__(self, address: str, authkey: bytes | str, chunk_size: int = 1024):
        host, port_str = address.split(":")
        self.host = host
        self.port = int(port_str)

        if isinstance(authkey, str):
            authkey = bytes.fromhex(authkey)

        assert isinstance(authkey, bytes), f"Invalid authkey type: {type(authkey)}"
        self.authkey = authkey
        self._conn = None
        self.chunk_size = chunk_size

    @property
    def address(self) -> tuple[str, int]:
        return (self.host, self.port)

    @property
    def conn(self) -> multiprocessing.connection.Connection:
        if self._conn is None:
            self._conn = multiprocessing.connection.Client(
                self.address, family="AF_INET", authkey=self.authkey
            )
            logger.info(f"Connected to {self.address}")
        return self._conn

    def send_message(self, msg_type: str, data, **kwargs) -> bool:
        """
        Send a structured message through the connection.

        If data is a long list, split it into chunks.
        """
        try:
            if isinstance(data, list) and len(data) > self.chunk_size:
                message_id = uuid.uuid4().hex
                total_chunks = math.ceil(len(data) / self.chunk_size)

                for idx in range(total_chunks):
                    part = data[idx * self.chunk_size : (idx + 1) * self.chunk_size]
                    msg = {
                        "type": msg_type,
                        "data": part,
                        "timestamp": time.time(),
                        "message_id": message_id,
                        "chunk_index": idx,
                        "total_chunks": total_chunks,
                        **kwargs,
                    }
                    self.conn.send(msg)
                logger.debug(
                    f"→ Streamed {len(data)} items in {total_chunks} chunks under id {message_id}"
                )
                return True

            # small or non-list payload: single shot
            msg = {
                "type": msg_type,
                "data": data,
                "timestamp": time.time(),
                **kwargs,
            }
            self.conn.send(msg)
            logger.debug(f"→ Sent single message: {msg_type}")
            return True

        except Exception as e:
            logger.error(f"Failed to send message: {e}", exc_info=True)
            return False

    @property
    def closed(self):
        return self.conn.closed

    @property
    def readable(self):
        return self.conn.readable

    @property
    def writable(self):
        return self.conn.writable

    def fileno(self):
        return self.conn.fileno()

    def recv(self):
        return self.conn.recv()

    def poll(self, timeout=0.0):
        return self.conn.poll(timeout)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        PROPOGATE = False
        SUPPRESS = True
        if exc_type:
            logger.error("Connection closed by parent")
            return PROPOGATE

        if self.conn is not None:
            try:
                logger.info("Closing worker-side connection.")
                self.conn.close()
            except Exception as e:
                logger.error(f"Error closing connection: {e}")
        return SUPPRESS


class WorkerController:
    """Wrap an AsyncEventEmitter in its own event loop

    Optionally uses QtAsyncio for better Qt integration if available.
    Set use_qtasyncio=True to enable Qt-native asyncio event loop.
    """

    def __init__(self, emitter_instance: AsyncEventEmitter, use_qtasyncio: bool = False):
        self.emitter = emitter_instance
        self.use_qtasyncio = use_qtasyncio and QTASYNCIO_ENABLED

        if self.use_qtasyncio:
            # Use Qt-native asyncio event loop
            logger.info("Using QtAsyncio event loop for worker")
            qt_set_event_loop_policy()
            self.loop = asyncio.new_event_loop()
        else:
            # Use standard asyncio event loop
            self.loop = asyncio.new_event_loop()

        self._thread = threading.Thread(target=self._run_loop, daemon=True)
        self._result = None
        self._started = False

    def _run_loop(self):
        asyncio.set_event_loop(self.loop)
        try:
            self._result = self.loop.run_until_complete(self.emitter.run())
        except Exception as e:
            logger.error(f"Exception in worker thread loop: {e}", exc_info=True)
            self._result = None

    def start(self):
        if self._started:
            logger.warning("Start called on an already started worker controller.")
            return
        self._thread.start()
        self._started = True

    def pause(self):
        if not self._started:
            logger.warning("Pause called before worker controller was started.")
            return
        logger.info("▶️  Pausing...")
        self.loop.call_soon_threadsafe(self.emitter.pause_evt.set)

    def resume(self):
        if not self._started:
            logger.warning("Resume called before worker controller was started.")
            return
        logger.info("▶️  Resuming...")
        self.loop.call_soon_threadsafe(self.emitter.pause_evt.clear)

    def stop(self):
        if not self._started:
            # Even if not started, set stop event for consistency if needed
            # self.loop.call_soon_threadsafe(self.emitter.stop_evt.set) # Maybe not necessary if loop never runs
            logger.info("🛑  Stopping (controller not started)...")
            if hasattr(
                self.emitter, "stop_evt"
            ):  # Emitter might not be fully initialized
                self.loop.call_soon_threadsafe(self.emitter.stop_evt.set)
            return
        logger.info("🛑  Stopping...")
        self.loop.call_soon_threadsafe(self.emitter.stop_evt.set)

    def join(self):
        if not self._started:
            logger.warning(
                "Join called before worker controller was started. Returning current result."
            )
            return self._result

        if self._thread.is_alive():
            self._thread.join()
        else:
            logger.info(
                "Worker thread was not alive when join was called (already finished or failed?)."
            )
        self._started = False  # Reset for potential restart, though typically not done.
        return self._result

    def set_log_level(self, level):
        # This assumes the logger used by emitter.run() is the global one or accessible
        # If emitter uses its own logger, it needs a method to set its level.
        # For now, we assume it affects the logger instance used within emitter.
        # A more robust solution might involve passing a logger instance or a config.
        logger.setLevel(level)
        if hasattr(self.emitter, "logger"):
            self.emitter.logger.setLevel(level)
        logger.info(f"Log level set to {level} for controller and its emitter.")


class WorkerBase(WorkerProtocol):
    """Base class for worker implementations."""

    def __init__(
        self,
        async_emitter_class: typing.Type[AsyncEventEmitter] | None = None,
        emitter_args: dict | None = None,
        process_chunk_fn: typing.Callable | None = None,
    ):
        self.async_emitter_class = async_emitter_class
        self.emitter_args = emitter_args or {}
        self.process_chunk_fn = process_chunk_fn  # User-provided chunk processor
        self.emitter_instance: AsyncEventEmitter | None = None
        self.controller: WorkerController | None = None
        self.conn: ConnectionContext | None = None

        self._commands = {
            "stop": self._handle_stop,
            "pause": self._handle_pause,
            "resume": self._handle_resume,
            "start": self._handle_start,  # Added start command
            "ping": self._handle_ping,  # Added ping command
            "set_log_level": self._handle_set_log_level,  # Added log level command
        }
        self.logger = get_logger(self.__class__.__name__)
        self._running = False  # Tracks if the process loop is active
        # self._paused is managed by the emitter's pause_evt now

    def setup(self, **kwargs):
        """Default setup: Initializes the AsyncEventEmitter if provided."""
        if self.async_emitter_class:
            # Pass process_chunk_fn to the emitter if it accepts it
            # This requires the AsyncEventEmitter subclass to handle it in its __init__ or a setter
            current_emitter_args = self.emitter_args.copy()
            if self.process_chunk_fn:
                current_emitter_args["process_chunk_fn"] = self.process_chunk_fn

            self.emitter_instance = self.async_emitter_class(**current_emitter_args)
            self._setup_default_event_handlers()
        else:
            self.logger.warning(
                "No async_emitter_class provided for WorkerBase. Task processing will not occur."
            )

    def _setup_default_event_handlers(self):
        """Set up default event handlers for the emitter instance."""
        if not self.emitter_instance:
            return

        @self.emitter_instance.on("run_started")
        def on_run_started():
            self.logger.info("▶️  Task starting")
            if self.conn:
                self.conn.send_message(
                    "progress", 0.0, status="running", stage="starting"
                )

        @self.emitter_instance.on("run_finished")
        def on_run_finished(results):
            # The structure of 'results' depends on the emitter's run() method.
            # For AntiDeobWorker, it's an IntervalSet.
            # We need a generic way to count/summarize or pass through.
            count = (
                len(results)
                if hasattr(results, "__len__")
                else (1 if results is not None else 0)
            )
            self.logger.info("✅ Task finished, processed %s items/results.", count)
            if self.conn:
                self.conn.send_message(
                    "progress",
                    0.95,  # Assuming this is a placeholder for actual progress
                    status="finalizing",
                    stage="task_complete",  # Generic stage
                    items_count=count,
                )

        @self.emitter_instance.on("stopped")
        def on_stopped():
            self.logger.info("🛑 Emitter shutting down")
            if self.conn:
                self.conn.send_message("status", "stopped", status="stopped")

        # Allow subclasses to add more specific handlers
        self.setup_custom_event_handlers()

    def setup_custom_event_handlers(self):
        """Subclasses can override this to add their own event handlers for the emitter."""
        pass

    async def cleanup(self):
        """Default cleanup: Shuts down the emitter and controller."""
        if self.controller:
            self.controller.stop()
            # Join should ideally happen, but might block. Consider if necessary or how to handle.
            # self.controller.join() # This might be problematic in an async cleanup

        if self.emitter_instance:
            await self.emitter_instance.shutdown()
        self.logger.info("Worker cleanup complete.")

    def handle_command(self, cmd: dict, conn: ConnectionContext) -> bool:
        """
        Handle standard commands. Override for custom commands.
        Returns True to continue processing, False to exit.
        """
        cmd_type = cmd.get("command")
        if cmd_type is None:
            return True
        handler = self._commands.get(cmd_type)
        if handler:
            return handler(cmd, conn)
        return True

    def _handle_stop(self, cmd, conn):
        """Handle stop command."""
        self.logger.info("Received stop command.")
        self._running = False  # Signal process loop to stop
        if self.controller:
            self.controller.stop()
        conn.send_message("status", "stopped", status="stopped")
        return False  # Exit process loop

    def _handle_pause(self, cmd, conn):
        """Handle pause command."""
        if self.controller:
            self.controller.pause()
            conn.send_message("status", "paused", status="paused")
        else:
            conn.send_message("error", "Not started", status="error")
        return True

    def _handle_resume(self, cmd, conn):
        """Handle resume command."""
        if self.controller:
            self.controller.resume()
            conn.send_message("status", "resumed", status="running")
        else:
            conn.send_message("error", "Not started", status="error")
        return True

    def _handle_start(self, cmd, conn):
        """Handle start command."""
        if not self.emitter_instance:
            self.logger.error("Cannot start: emitter_instance not initialized.")
            conn.send_message(
                "error", "Worker not properly configured (no emitter)", status="error"
            )
            return True  # Stay in command loop

        if self.controller and self.controller._started:
            self.logger.warning("Start command received, but already started.")
            conn.send_message("status", "already_running", status="running")
            return True

        self.logger.info(
            "Received start command. Initializing and starting controller."
        )
        self.controller = WorkerController(self.emitter_instance)
        self.controller.start()
        conn.send_message("status", "started", status="running")
        return True

    def _handle_ping(self, cmd, conn):
        """Handle ping command."""
        conn.send_message(
            "status", "pong", status="running" if self._running else "idle"
        )
        return True

    def _handle_set_log_level(self, cmd, conn):
        """Handle set_log_level command."""
        level = cmd.get("level")
        if level is None:
            self.logger.error("Log level is required for set_log_level command.")
            conn.send_message("error", "Log level not specified", status="error")
            return True

        try:
            # self.logger.setLevel(level) # WorkerBase logger
            # if self.emitter_instance and hasattr(self.emitter_instance, 'logger'):
            #     self.emitter_instance.logger.setLevel(level)
            if self.controller:
                self.controller.set_log_level(level)  # Propagates to emitter logger too
            else:  # Set global logger if controller not up
                logger.setLevel(level)

            self.logger.info(
                f"Log level set to {level}"
            )  # Changed from logger.info to self.logger.info
            conn.send_message("status", f"log_level_set:{level}", status="running")
        except Exception as e:
            self.logger.error(f"Failed to set log level: {e}", exc_info=True)
            conn.send_message(
                "error", f"Failed to set log level: {str(e)}", status="error"
            )
        return True

    def process(self, connection: ConnectionContext, **kwargs):
        """Main processing loop - implement in subclasses."""
        # raise NotImplementedError("Subclasses must implement process()")
        self.conn = connection
        self._running = True

        # Send initial ready message
        connection.send_message("status", "connected", status="ready")
        self.logger.info("Worker connected, awaiting commands...")

        try:
            while self._running:
                try:
                    # Poll for messages, timeout allows checking self._running
                    if not connection.closed and not connection.poll(timeout=0.5):
                        # self.logger.debug("No message, continuing poll.")
                        if not self._running:  # Check if stop was called during poll
                            break
                        continue
                    if connection.closed:  # check before recv
                        self.logger.error(
                            "Connection closed by parent (detected before recv)."
                        )
                        self._running = False
                        break
                    cmd = connection.recv()
                    self.logger.debug(f"← Received command: {cmd}")
                except EOFError:
                    self.logger.error("Connection closed by parent (EOFError).")
                    self._running = False  # Ensure loop termination
                    break
                except (ConnectionResetError, BrokenPipeError) as e:
                    self.logger.error(f"Connection error: {e}")
                    self._running = False  # Ensure loop termination
                    break
                except Exception as e:
                    self.logger.error(
                        f"Unexpected error receiving command: {e}", exc_info=True
                    )
                    # Optionally send error message back if connection is still usable
                    # connection.send_message("error", f"Error receiving command: {str(e)}", status="error")
                    # Depending on error, might need to stop.
                    # For now, continue to allow graceful shutdown via 'stop' command.
                    continue

                if (
                    not self._running
                ):  # check after recv, if stop was part of the message processing
                    break

                # Handle command
                if isinstance(cmd, dict):
                    if not self.handle_command(cmd, connection):
                        self._running = False  # Command handler requested exit
                        break
                elif cmd is None:  # Typically means connection closed
                    self.logger.info("Received None, likely connection closed.")
                    self._running = False
                    break
                else:
                    self.logger.warning(
                        f"Received unexpected command type: {type(cmd)}"
                    )
                    connection.send_message(
                        "error",
                        f"Expected dict command, got {type(cmd)}",
                        status="error",
                    )
            # Loop exited, ensure controller stops if it was started
            if self.controller:
                self.logger.info(
                    "Process loop finished. Ensuring controller is stopped."
                )
                self.controller.stop()

        finally:
            self.logger.info("Process loop finalizing.")
            if self.controller:
                self.logger.info("Waiting for task controller to join...")
                results = self.controller.join()
                if results is not None:
                    self.send_results(connection, results)
                else:
                    self.logger.info(
                        "No results from controller or task did not complete successfully."
                    )
            else:
                self.logger.info("Controller was not active or not used.")

            self.conn = None  # Clear connection reference
            self.logger.info("Worker process loop ended.")

    def send_results(self, connection: ConnectionContext, results_data):
        """Formats and sends results. Subclasses can override for custom formatting."""
        # Default: assume results_data is list-like or an IntervalSet
        if isinstance(results_data, IntervalSet):
            as_json = [
                {"address": s, "length": e - s, "end": e}
                for s, e in results_data.as_tuples()
            ]
        elif isinstance(results_data, list):
            # Assume it's already in a suitable JSON-serializable list format
            as_json = results_data
        elif results_data is not None:
            # Attempt to make it a list if it's a single non-list item
            as_json = [results_data]
        else:
            as_json = []

        if as_json:
            self.logger.info(f"Sending {len(as_json)} results...")
            connection.send_message(
                "status", "sending_results", status="sending_results"
            )
            # Consider making sleep configurable or removing if IDA side handles UI updates better
            # time.sleep(1) # Shorter sleep or make IDA side robust
            connection.send_message(
                "result", as_json, status="success", count=len(as_json)
            )
            connection.send_message("status", "results_sent", status="results_sent")
        else:
            self.logger.info("No results to send or results were empty.")
            # Send a success message even if results are empty, or a specific no_results status
            connection.send_message(
                "result", [], status="success", count=0, note="No results generated"
            )

    # Subclasses might need to implement their own run_async_task or similar
    # if the AsyncEventEmitter pattern is not sufficient.



"""
Decorators for simplifying async task execution.

Provides simple decorator-based API for running CPU-intensive tasks
in the background without blocking.
"""

import functools
from typing import Callable, Optional, Any
from concurrent.futures import Future

from .qt_compat import QT_AVAILABLE

if QT_AVAILABLE:
    from .qtasyncio import ProcessPoolExecutor, ThreadExecutor

# Global executor instance (created on first use)
_global_executor = None


def get_global_executor(max_workers: Optional[int] = None):
    """Get or create the global background task executor."""
    global _global_executor

    if _global_executor is None:
        if QT_AVAILABLE:
            _global_executor = ProcessPoolExecutor(max_workers=max_workers)
        else:
            raise RuntimeError("Qt is not available - cannot create executor")

    return _global_executor


def background_task(
    func: Optional[Callable] = None,
    *,
    max_workers: Optional[int] = None,
    on_complete: Optional[Callable[[Any], None]] = None,
    on_error: Optional[Callable[[Exception], None]] = None,
    on_progress: Optional[Callable[[int, str], None]] = None,
    timeout: Optional[float] = None,
    executor_type: str = 'thread'  # 'thread' (default) or 'process'
):
    """
    Decorator to run a function as a background task without blocking.

    Transforms a regular function into one that runs in a separate process/thread
    and returns a Future immediately.

    Args:
        func: The function to decorate (auto-filled when used without parens)
        max_workers: Maximum number of parallel workers (default: CPU count)
        on_complete: Callback when task completes, receives result
        on_error: Callback when task fails, receives exception
        on_progress: Callback for progress updates (progress%, message)
        timeout: Optional timeout in seconds
        executor_type: 'process' for CPU-bound, 'thread' for I/O-bound

    Returns:
        Decorated function that returns a Future when called

    Examples:
        Basic usage:
        >>> @background_task
        ... def slow_function(x):
        ...     return x * 2
        ...
        >>> future = slow_function(5)  # Returns immediately
        >>> result = future.result()   # Get result when needed

        With callback:
        >>> @background_task(on_complete=lambda r: print(f"Done: {r}"))
        ... def slow_function(x):
        ...     return x * 2
        ...
        >>> slow_function(5)  # Callback fires when done

        With progress:
        >>> @background_task(
        ...     on_progress=lambda p, m: print(f"[{p}%] {m}"),
        ...     on_complete=lambda r: print(f"Result: {r}")
        ... )
        ... def slow_function(x, progress_callback=None):
        ...     if progress_callback:
        ...         progress_callback(50, "Halfway done")
        ...     return x * 2

        Multiple parallel tasks:
        >>> @background_task(max_workers=4)
        ... def process_item(item):
        ...     return expensive_computation(item)
        ...
        >>> futures = [process_item(x) for x in range(100)]
        >>> results = [f.result() for f in futures]
    """

    def decorator(fn: Callable) -> Callable:
        """The actual decorator function."""

        # Create or get executor
        if executor_type == 'thread' and QT_AVAILABLE:
            from .qtasyncio import ThreadExecutor
            executor = ThreadExecutor(max_workers=max_workers)
        else:
            executor = get_global_executor(max_workers=max_workers)

        # Connect callbacks to signals if provided
        if on_complete:
            def handle_complete(future: Future):
                try:
                    result = future.result(timeout=timeout)
                    on_complete(result)
                except Exception:
                    pass  # Error handled by on_error if registered

            executor.signals.task_completed.connect(handle_complete)

        if on_error:
            def handle_error(future: Future, exception: Exception):
                on_error(exception)

            executor.signals.task_failed.connect(handle_error)

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            """Wrapper that submits the task to background executor."""

            # Inject progress callback if the function accepts it
            if on_progress:
                import inspect
                sig = inspect.signature(fn)
                if 'progress_callback' in sig.parameters:
                    kwargs['progress_callback'] = on_progress

            # Submit to executor (returns immediately)
            future = executor.submit(fn, *args, **kwargs)

            return future

        # Attach executor for manual control if needed
        wrapper.executor = executor

        return wrapper

    # Handle both @background_task and @background_task(...)
    if func is None:
        # Called with arguments: @background_task(max_workers=4)
        return decorator
    else:
        # Called without arguments: @background_task
        return decorator(func)


def cpu_task(
    func: Optional[Callable] = None,
    *,
    max_workers: Optional[int] = None,
    on_complete: Optional[Callable] = None,
    on_error: Optional[Callable] = None,
):
    """
    Simplified decorator for CPU-intensive tasks.

    Uses ThreadExecutor by default for compatibility. For truly parallel CPU work
    (bypassing GIL), use ProcessPoolExecutor directly.

    Note: In IDA, most CPU work calls IDA SDK functions which release the GIL,
    so threads work well. Use this for pattern analysis, signature generation, etc.

    Example:
        >>> @cpu_task(on_complete=lambda r: print(f"Result: {r}"))
        ... def find_signature(data):
        ...     # CPU-intensive work here
        ...     return analyze_pattern(data)
        ...
        >>> future = find_signature(binary_data)  # Runs in background thread
    """
    return background_task(
        func,
        max_workers=max_workers,
        on_complete=on_complete,
        on_error=on_error,
        executor_type='thread'
    )


def io_task(
    func: Optional[Callable] = None,
    *,
    max_workers: Optional[int] = None,
    on_complete: Optional[Callable] = None,
    on_error: Optional[Callable] = None,
):
    """
    Simplified decorator for I/O-bound tasks.

    Alias for @background_task with executor_type='thread'.
    Use for I/O-bound work (network requests, file operations, database queries, etc.)

    Example:
        >>> @io_task(on_complete=lambda r: print(f"Downloaded: {r}"))
        ... def fetch_data(url):
        ...     return requests.get(url).json()
        ...
        >>> future = fetch_data("https://api.example.com/data")
    """
    return background_task(
        func,
        max_workers=max_workers,
        on_complete=on_complete,
        on_error=on_error,
        executor_type='thread'
    )


def parallel(max_workers: int = None):
    """
    Decorator for functions that should run multiple instances in parallel.

    Convenience decorator that sets up parallel execution with sensible defaults.
    Uses threads for compatibility - for true parallel CPU work, use ProcessPoolExecutor.

    Args:
        max_workers: Number of parallel workers (default: CPU count)

    Example:
        >>> @parallel(max_workers=8)
        ... def process_function(address):
        ...     # Analyze function at address
        ...     return analysis_result
        ...
        >>> # Process 100 functions in parallel across 8 workers
        >>> futures = [process_function(addr) for addr in function_addresses]
        >>> results = [f.result() for f in futures]
    """
    return background_task(max_workers=max_workers, executor_type='thread')


def shared_memory_task(
    func: Optional[Callable] = None,
    *,
    num_chunks: int = 8,
    max_workers: Optional[int] = None,
    on_complete: Optional[Callable[[list], None]] = None,
    on_error: Optional[Callable[[Exception], None]] = None,
):
    """
    Decorator for processing large data using shared memory across multiple processes.

    Handles all the shared memory complexity:
    - Creates shared memory segment
    - Copies data once
    - Calculates chunk boundaries
    - Creates workers that attach to shared memory
    - Collects results
    - Cleanup (memoryview, close, unlink)

    User just writes the chunk processing logic!

    Args:
        func: The chunk processing function (auto-filled when used without parens)
        num_chunks: Number of chunks to split data into (default: 8)
        max_workers: Maximum number of parallel workers (default: num_chunks)
        on_complete: Callback when all chunks complete, receives list of results
        on_error: Callback when any chunk fails, receives exception

    The decorated function receives:
        chunk_data: memoryview of this chunk's data
        chunk_id: 0-based chunk index
        total_chunks: total number of chunks

    Returns:
        Decorated function that takes full data and returns list of all chunk results

    Example:
        >>> @shared_memory_task(num_chunks=8)
        ... def analyze_chunk(chunk_data, chunk_id, total_chunks):
        ...     # Process this chunk
        ...     signatures = find_patterns(chunk_data)
        ...     return {'chunk': chunk_id, 'sigs': signatures}
        ...
        >>> # Usage - just pass the full data!
        >>> results = analyze_chunk(binary_data)  # Returns list of 8 results
        >>> # ida-taskr handles all shared memory complexity

    Real-world IDA example:
        >>> @shared_memory_task(num_chunks=16)
        ... def find_signatures(chunk_data, chunk_id, total_chunks):
        ...     signatures = []
        ...     for i in range(len(chunk_data)):
        ...         if is_interesting_pattern(chunk_data[i:i+16]):
        ...             signatures.append(bytes(chunk_data[i:i+16]))
        ...     return signatures
        ...
        >>> # Get binary data from IDA
        >>> binary_data = ida_bytes.get_bytes(start_ea, size)
        >>> # Process in parallel - shared memory handles 8MB+ efficiently
        >>> all_signatures = find_signatures(binary_data)
    """
    import multiprocessing.shared_memory as shm_module

    def decorator(fn: Callable) -> Callable:
        """The actual decorator function."""

        @functools.wraps(fn)
        def wrapper(data: bytes) -> list:
            """
            Wrapper that handles all shared memory setup/teardown.

            Args:
                data: Full binary data to process

            Returns:
                List of results from all chunks
            """
            if not QT_AVAILABLE:
                raise RuntimeError("Qt is not available - cannot use shared memory tasks")

            # Create shared memory segment
            shm = shm_module.SharedMemory(create=True, size=len(data))

            try:
                # Copy data into shared memory (ONCE!)
                shm.buf[:len(data)] = data

                # Calculate chunk boundaries
                chunk_size = len(data) // num_chunks
                chunks_info = []

                for i in range(num_chunks):
                    start = i * chunk_size
                    # Last chunk gets any remainder
                    end = start + chunk_size if i < num_chunks - 1 else len(data)
                    chunks_info.append((start, end, i))

                # Create executor
                workers = max_workers if max_workers else num_chunks
                executor = ProcessPoolExecutor(max_workers=workers)

                try:
                    # Submit all chunks
                    futures = []
                    for start, end, chunk_id in chunks_info:
                        future = executor.submit(
                            _shared_memory_worker,
                            fn,
                            shm.name,
                            start,
                            end,
                            chunk_id,
                            num_chunks
                        )
                        futures.append(future)

                    # Collect results
                    results = []
                    for future in futures:
                        try:
                            result = future.result()
                            results.append(result)
                        except Exception as e:
                            if on_error:
                                on_error(e)
                            raise

                    # Call completion callback if provided
                    if on_complete:
                        on_complete(results)

                    return results

                finally:
                    executor.shutdown(wait=True)

            finally:
                # Cleanup shared memory
                shm.close()
                shm.unlink()

        return wrapper

    # Handle both @shared_memory_task and @shared_memory_task(...)
    if func is None:
        # Called with arguments: @shared_memory_task(num_chunks=16)
        return decorator
    else:
        # Called without arguments: @shared_memory_task
        return decorator(func)


def _shared_memory_worker(fn: Callable, shm_name: str, start: int, end: int, chunk_id: int, total_chunks: int):
    """
    Worker function that attaches to shared memory and processes a chunk.

    This needs to be a module-level function so it can be pickled for ProcessPoolExecutor.

    Args:
        fn: The user's chunk processing function
        shm_name: Name of the shared memory segment
        start: Start offset in shared memory
        end: End offset in shared memory
        chunk_id: This chunk's ID
        total_chunks: Total number of chunks

    Returns:
        Result from user's chunk processing function
    """
    import multiprocessing.shared_memory as shm_module

    # Attach to existing shared memory
    shm = shm_module.SharedMemory(name=shm_name)

    try:
        # Get view of this chunk (no data copying!)
        chunk_data = memoryview(shm.buf)[start:end]

        # Call user's function with chunk data
        result = fn(chunk_data, chunk_id, total_chunks)

        return result

    finally:
        # CRITICAL: Delete memoryview before closing shared memory
        del chunk_data
        shm.close()


# Export all decorators
__all__ = [
    'background_task',
    'cpu_task',
    'io_task',
    'parallel',
    'shared_memory_task',
    'get_global_executor',
]



"""Helper utilities for the worker manager."""

import functools
import logging
import multiprocessing
import os
import pathlib
import stat
import sys


def is_ida():
    """
    Crude check to see if running inside IDA.
    Returns True if running inside IDA Pro, else False.

    we do not use this check because there's a possibility that
    we're running inside a headless IDA mode and that uses an
    `idapro` library which will make this succeed.

    Even though that is technically "running" in IDA, we can
    still use the python interpreter that's executing the script.
    The reason we want to know if we're in the IDA application is
    because we want to find the python interpreter that IDA is using
    to execute the script and to not start a new IDA instance.

    Maybe this function can be called `is_ida_application` or something.

        try:
            import idaapi  # noqa
            return True
        except ImportError:
            return False
    """
    # Check if sys.executable is an IDA executable
    exec_name = pathlib.Path(sys.executable).name.lower()
    if exec_name.startswith(("ida", "idat", "idaw", "idag")):
        return True

    # Also check if stdout is IDAPythonStdOut (when running scripts via -S flag)
    # This handles the case where IDA runs scripts with -S"script.py"
    if type(sys.stdout).__name__ == "IDAPythonStdOut":
        return True

    return False


# on windows, we need to set the encoding to utf-8 because it defaults to cp1252
# which does not support the emoji characters used in the logging
# or really any non-ascii characters
if not is_ida():
    # this works in non IDA and for python 3.7+
    sys.stdout.reconfigure(encoding="utf-8")  # type: ignore
else:
    # IDA wraps sys.stdout and does not expose the `reconfigure` method
    # so we need to set the encoding manually
    sys.stdout.encoding = "utf-8"  # type: ignore


def configure_logging(
    log,
    level=logging.INFO,
    handler_filters=None,
    fmt_str="[%(name)s:%(levelname)s:%(process)d:%(threadName)s] @ %(asctime)s %(message)s",
):
    """Configure logging with proper formatting and filters."""
    log.propagate = False
    log.setLevel(level)
    formatter = logging.Formatter(fmt_str)
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(formatter)
    handler.setLevel(level)

    if handler_filters is not None:
        for _filter in handler_filters:
            handler.addFilter(_filter)

    for handler in log.handlers[:]:
        log.removeHandler(handler)
        handler.close()

    if not log.handlers:
        log.addHandler(handler)


def get_logger(name=None, configurer=None, log_level=logging.INFO, custom_logger=None):
    """Get a configured logger instance."""
    if custom_logger:
        return custom_logger
    if not configurer:
        configurer = functools.partial(configure_logging, level=log_level)
    prefix = "ida." if is_ida() else "worker."
    name = name or f"{prefix}{__name__}"
    logger = logging.getLogger(name)
    configurer(logger)
    return logger


class MultiprocessingHelper:
    """
    Static helper class for multiprocessing context and Python interpreter discovery.
    """

    @staticmethod
    @functools.lru_cache(maxsize=1)
    def get_python_interpreter():
        """
        Gets the path to a suitable Python interpreter.
        Ensures we find a standalone Python executable.
        >>> import pathlib, sys, stat
        >>> interp: pathlib.Path = MultiprocessingHelper.get_python_interpreter()
        ...
        >>>
        """
        logger = get_logger()

        if (
            hasattr(sys, "_base_executable")  # type: ignore
            and sys._base_executable  # type: ignore
            and "python" in pathlib.Path(sys._base_executable).name.lower()  # type: ignore
        ):
            logger.debug(f"Using _base_executable: {sys._base_executable}")  # type: ignore
            return pathlib.Path(sys._base_executable)  # type: ignore

        base_paths = [
            sys.prefix,
            sys.exec_prefix,
            sys.executable,
        ]
        exe_suffix = ".exe" if os.name == "nt" else ""
        python_name = f"python{exe_suffix}"

        def base_dirs():
            for dirname in map(pathlib.Path, base_paths):
                yield dirname
                yield dirname.parent
                yield dirname.parent.parent

        for dirname in base_dirs():
            for basename in ["", "bin", "python"]:
                interp_path = dirname / basename / python_name
                if not interp_path.exists():
                    continue

                if not (interp_path.is_file() or interp_path.is_symlink()):
                    continue

                st_mode = interp_path.stat().st_mode
                if not st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
                    continue

                logger.debug(f"Found Python interpreter at: {interp_path}")
                return interp_path

        logger.warning(
            "Could not determine Python interpreter path, falling back to 'python' in PATH."
        )
        return pathlib.Path("python")

    @staticmethod
    def set_multiprocessing_context():
        """
        Sets up the multiprocessing context to use 'spawn' and sets the Python executable.
        """
        current_method = multiprocessing.get_start_method(allow_none=True)
        if current_method != "spawn":
            multiprocessing.set_start_method("spawn", force=True)
        multiprocessing.set_executable(
            str(MultiprocessingHelper.get_python_interpreter())
        )


# Initialize multiprocessing context
MultiprocessingHelper.set_multiprocessing_context()



"""Qt compatibility layer for PyQt5 and PySide6.

This module provides a unified API for Qt signals and slots across different
Qt bindings used by IDA Pro and standalone environments.
"""

# Try to import Qt frameworks in order of preference
QT_API = None
QT_AVAILABLE = False
QtCore = None
Signal = None
Slot = None
QT_VERSION = None
QProcessEnvironment = None

# Try PySide6 first (IDA Pro 9.2+, modern official Qt bindings)
try:
    from PySide6 import QtCore
    from PySide6.QtCore import Signal, Slot

    QT_API = "PySide6"
    QT_VERSION = QtCore.__version__
    QProcessEnvironment = QtCore.QProcessEnvironment
    QT_AVAILABLE = True
except (ImportError, NotImplementedError):
    pass

# Try PyQt5 (IDA Pro 9.1 and earlier)
if QT_API is None:
    try:
        from PyQt5 import QtCore
        from PyQt5.QtCore import pyqtSignal as Signal, pyqtSlot as Slot

        QT_API = "PyQt5"
        QT_VERSION = QtCore.PYQT_VERSION_STR
        QProcessEnvironment = QtCore.QProcessEnvironment
        QT_AVAILABLE = True
    except (ImportError, NotImplementedError):
        pass

# If no Qt found, create mock classes for import compatibility
if QT_API is None:

    class QtCore:  # type: ignore
        """Mock QtCore module when Qt is not available."""

        class QThread:
            def __init__(self, *args, **kwargs):
                raise ImportError(
                    "Qt is not available. Cannot use WorkerLauncher without Qt."
                )

        class QObject:
            def __init__(self, *args, **kwargs):
                raise ImportError(
                    "Qt is not available. Cannot use WorkerLauncher without Qt."
                )

        class QProcess:
            # Process error enum values (mock)
            class ProcessError:
                FailedToStart = 0
                Crashed = 1
                Timedout = 2
                WriteError = 4
                ReadError = 3
                UnknownError = 5

            # Process state enum values (mock)
            class ProcessState:
                NotRunning = 0
                Starting = 1
                Running = 2

            # Direct enum access (compatibility)
            FailedToStart = 0
            Crashed = 1
            Timedout = 2
            WriteError = 4
            ReadError = 3
            UnknownError = 5
            NotRunning = 0
            Starting = 1
            Running = 2
            NormalExit = 0
            CrashExit = 1

            def __init__(self, *args, **kwargs):
                raise ImportError(
                    "Qt is not available. Cannot use WorkerLauncher without Qt."
                )

        class QSocketNotifier:
            Read = 1
            Write = 2

            def __init__(self, *args, **kwargs):
                raise ImportError(
                    "Qt is not available. Cannot use WorkerLauncher without Qt."
                )

    # Dummy signal for type hints
    Signal = lambda *args: None  # type: ignore
    Slot = lambda *args: None  # type: ignore

    class QProcessEnvironment:  # type: ignore
        @staticmethod
        def systemEnvironment():
            raise ImportError(
                "Qt is not available. Cannot use WorkerLauncher without Qt."
            )


def get_qt_api():
    """Return the name of the Qt API being used, or None if Qt is not available."""
    return QT_API


def get_qt_version():
    """Return the version of the Qt framework being used, or None if Qt is not available."""
    return QT_VERSION


# QtAsyncio module availability
QT_ASYNCIO_AVAILABLE = False
qtasyncio = None

if QT_AVAILABLE:
    try:
        from . import qtasyncio
        QT_ASYNCIO_AVAILABLE = True
    except ImportError:
        pass


__all__ = [
    "QtCore",
    "Signal",
    "Slot",
    "QT_API",
    "QT_VERSION",
    "QT_AVAILABLE",
    "QProcessEnvironment",
    "get_qt_api",
    "get_qt_version",
    "QT_ASYNCIO_AVAILABLE",
    "qtasyncio",
]



# QtAsyncio + ThreadExecutor + Worker utilities (single-file, typed, PySide6-compatible)
# Commit: qtproject/pyside-pyside-setup@072ffd057a29a694a0ad91894736bb4d0a88738e + extras
# Includes: asyncio integration, ThreadExecutor, Task/FutureWatcher, thread_worker utils

from __future__ import annotations

import asyncio
import atexit
import concurrent.futures
import inspect
import logging
import sys
import threading
import time
import warnings
import weakref
from contextlib import contextmanager
from functools import partial, wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Generator,
    List,
    Optional,
    ParamSpec,
    Sequence,
    Tuple,
    TypeVar,
    overload,
)

from .qt_compat import QtCore, Signal, Slot, QT_AVAILABLE

if not QT_AVAILABLE:
    raise ImportError(
        "QtAsyncio module requires Qt (PyQt5 or PySide6) to be available. "
        "Please install PyQt5 or PySide6."
    )

# Import Qt components we need
QCoreApplication = QtCore.QCoreApplication
QEvent = QtCore.QEvent
QEventLoop = QtCore.QEventLoop
QMetaObject = QtCore.QMetaObject
QObject = QtCore.QObject
QSemaphore = QtCore.QSemaphore
QThread = QtCore.QThread
QThreadPool = QtCore.QThreadPool
QTimer = QtCore.QTimer
Qt = QtCore.Qt

# Try to import Q_ARG and QRunnable with compatibility handling
try:
    from PySide6.QtCore import Q_ARG, QRunnable
except ImportError:
    try:
        from PyQt5.QtCore import Q_ARG, QRunnable
    except ImportError:
        # Fallback if neither is available
        Q_ARG = None
        QRunnable = object  # type: ignore

if sys.version_info >= (3, 11):
    from typing import Self
else:
    Self = Any

# Type variable for Future
Future = concurrent.futures.Future

# --------------------------------------------------------------------------- #
# Asyncio integration (original QtAsyncio module)
# --------------------------------------------------------------------------- #
class QAsyncioEventLoop(QEventLoop):
    def __init__(self, asyncio_loop: asyncio.AbstractEventLoop, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._asyncio_loop = asyncio_loop

    def processEvents(self, flags=None):
        if flags is None:
            flags = QEventLoop.AllEvents
        self._asyncio_loop.run_until_complete(self._process_events(flags))

    async def _process_events(self, flags):
        while self._asyncio_loop._ready:
            await asyncio.sleep(0)
        super().processEvents(flags)
        await asyncio.sleep(0)


class QAsyncioEventLoopPolicy(
    asyncio.DefaultEventLoopPolicy
    if sys.platform != "win32"
    else asyncio.WindowsProactorEventLoopPolicy
):
    def new_event_loop(self):
        qt_app = QCoreApplication.instance() or QCoreApplication(sys.argv)
        asyncio_loop = super().new_event_loop()
        qt_loop = QAsyncioEventLoop(asyncio_loop, qt_app)

        def _wakeup_qt():
            qt_loop.wakeUp()

        asyncio_loop.call_soon(_wakeup_qt)
        timer = QTimer(qt_app)
        timer.timeout.connect(lambda: asyncio_loop.run_until_complete(asyncio.sleep(0)))
        timer.start(10)
        return asyncio_loop


def run(coroutine_or_future):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coroutine_or_future)


def set_event_loop_policy():
    """Set the Qt-compatible asyncio event loop policy."""
    if not isinstance(asyncio.get_event_loop_policy(), QAsyncioEventLoopPolicy):
        asyncio.set_event_loop_policy(QAsyncioEventLoopPolicy())


# --------------------------------------------------------------------------- #
# ThreadExecutor + Task + FutureWatcher + utilities
# --------------------------------------------------------------------------- #
@contextmanager
def locked(mutex):
    mutex.lock()
    try:
        yield
    finally:
        mutex.unlock()


class _TaskDepotThread(QThread):
    _lock = threading.Lock()
    _instance: Optional["_TaskDepotThread"] = None

    def __new__(cls):
        if _TaskDepotThread._instance is not None:
            raise RuntimeError("Already exists")
        return super().__new__(cls)

    def __init__(self):
        super().__init__()
        self.start()
        self.moveToThread(self)
        atexit.register(self._cleanup)

    def _cleanup(self):
        self.quit()
        self.wait()

    @staticmethod
    def instance() -> "_TaskDepotThread":
        with _TaskDepotThread._lock:
            if _TaskDepotThread._instance is None:
                _TaskDepotThread._instance = _TaskDepotThread()
            return _TaskDepotThread._instance

    @Slot(object, object)
    def transfer(self, obj: QObject, thread: QThread):
        assert obj.thread() is self
        assert QThread.currentThread() is self
        obj.moveToThread(thread)


class _TaskRunnable(QRunnable):
    def __init__(self, future, task, args, kwargs):
        super().__init__()
        self.future = future
        self.task = task
        self.args = args
        self.kwargs = kwargs
        self.eventLoop: Optional[QEventLoop] = None

    def run(self):
        self.eventLoop = QEventLoop()
        self.eventLoop.processEvents()

        if Q_ARG is not None:
            QMetaObject.invokeMethod(
                self.task.thread(),
                "transfer",
                Qt.BlockingQueuedConnection,
                Q_ARG(object, self.task),
                Q_ARG(object, QThread.currentThread()),
            )

        self.eventLoop.processEvents()
        self.task.start()
        self.task.finished.connect(self.eventLoop.quit)
        self.task.cancelled.connect(self.eventLoop.quit)
        self.eventLoop.exec_()


class FutureRunnable(QRunnable):
    def __init__(self, future, func, args, kwargs):
        super().__init__()
        self.future = future
        self.task = (func, args, kwargs)

    def run(self):
        try:
            if not self.future.set_running_or_notify_cancel():
                return
            func, args, kwargs = self.task
            result = func(*args, **kwargs)
            self.future.set_result(result)
        except BaseException as ex:
            self.future.set_exception(ex)
        except BaseException:
            logging.getLogger(__name__).critical("Exception in worker thread.", exc_info=True)


class ThreadPoolExecutorSignals(QObject):
    """Qt signals for ThreadExecutor events."""
    task_submitted = Signal(object)  # future
    task_completed = Signal(object)  # future
    task_failed = Signal(object, object)  # future, exception
    pool_shutdown = Signal()


class ThreadExecutor(QObject, concurrent.futures.Executor):
    """
    A ThreadExecutor that provides concurrent.futures.ThreadPoolExecutor-like API
    using Qt's QThreadPool with optional Qt signal integration.

    Usage:
        executor = ThreadExecutor()

        # Standard concurrent.futures API
        future = executor.submit(task, arg1, arg2)
        result = future.result()

        # Or with Qt signals
        executor.signals.task_completed.connect(on_task_done)
        future = executor.submit(task, arg1, arg2)

        executor.shutdown(wait=True)
    """

    def __init__(
        self,
        parent: Optional[QObject] = None,
        threadPool: Optional[QThreadPool] = None,
        max_workers: Optional[int] = None,
    ):
        super().__init__(parent)
        self.signals = ThreadPoolExecutorSignals()
        self._threadPool = threadPool or QThreadPool.globalInstance()

        # Set max thread count if specified
        if max_workers is not None:
            self._threadPool.setMaxThreadCount(max_workers)

        self._max_workers = max_workers or self._threadPool.maxThreadCount()
        self._depot_thread: Optional[_TaskDepotThread] = None
        self._futures: List[Future] = []
        self._shutdown = False
        self._state_lock = threading.Lock()

    @property
    def max_workers(self) -> int:
        """Return the maximum number of worker threads."""
        return self._max_workers

    def _get_depot_thread(self) -> _TaskDepotThread:
        if self._depot_thread is None:
            self._depot_thread = _TaskDepotThread.instance()
        return self._depot_thread

    def submit(self, func, *args, **kwargs) -> Future:
        """
        Submit a callable to be executed in a worker thread.

        Args:
            func: A callable to execute
            *args: Positional arguments for the callable
            **kwargs: Keyword arguments for the callable

        Returns:
            A Future representing the pending execution

        Raises:
            RuntimeError: If the executor has been shut down
        """
        with self._state_lock:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown.")
            if isinstance(func, Task):
                warnings.warn("Use `submit_task` to run `Task`s", DeprecationWarning, stacklevel=2)
                f, runnable = self.__make_task_runnable(func)
            else:
                f = Future()
                runnable = FutureRunnable(f, func, args, kwargs)
            self._futures.append(f)
            f.add_done_callback(self._future_done)
            self._threadPool.start(runnable)

            # Emit task_submitted signal
            try:
                self.signals.task_submitted.emit(f)
            except RuntimeError:
                pass  # Qt object may have been deleted

            return f

    def __make_task_runnable(self, task: Task) -> Tuple[Future, _TaskRunnable]:
        if task.thread() is not QThread.currentThread():
            raise ValueError("Can only submit Tasks from its own thread.")
        if task.parent() is not None:
            raise ValueError("Cannot submit Tasks with a parent.")
        task.moveToThread(self._get_depot_thread())
        f = task.future()
        runnable = _TaskRunnable(f, task, (), {})
        return f, runnable

    def map(
        self,
        fn: Callable,
        *iterables,
        timeout: Optional[float] = None,
        chunksize: int = 1,
    ):
        """
        Map a function over iterables, executing in parallel threads.

        Args:
            fn: A callable
            *iterables: Iterables of arguments
            timeout: Maximum time to wait for results
            chunksize: Ignored (for API compatibility with ProcessPoolExecutor)

        Returns:
            Iterator of results in the same order as the input iterables
        """
        with self._state_lock:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown.")

        # Submit all tasks
        futures = [self.submit(fn, *args) for args in zip(*iterables)]

        # Yield results in order
        for future in futures:
            yield future.result(timeout=timeout)

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False):
        """
        Shutdown the executor.

        Args:
            wait: If True, wait for all pending futures to complete
            cancel_futures: If True, cancel all pending futures (best effort)
        """
        with self._state_lock:
            self._shutdown = True
            futures = list(self._futures)

        if cancel_futures:
            for f in futures:
                f.cancel()

        if wait:
            concurrent.futures.wait(futures)

        try:
            self.signals.pool_shutdown.emit()
        except RuntimeError:
            pass  # Qt object may have been deleted

    def _future_done(self, future: Future):
        """Called when a future completes - emits appropriate Qt signals."""
        with self._state_lock:
            if future in self._futures:
                self._futures.remove(future)

        try:
            exc = future.exception()
            if exc is not None:
                self.signals.task_failed.emit(future, exc)
            else:
                self.signals.task_completed.emit(future)
        except (concurrent.futures.CancelledError, RuntimeError):
            pass  # Future was cancelled or Qt object deleted

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False


class Task(QObject):
    started = Signal()
    finished = Signal()
    cancelled = Signal()
    resultReady = Signal(object)
    exceptionReady = Signal(Exception)

    __ExecuteCall = QEvent.registerEventType()

    def __init__(self, parent: Optional[QObject] = None, function: Optional[Callable] = None):
        super().__init__(parent)
        warnings.warn("`Task` has been deprecated", PendingDeprecationWarning, stacklevel=2)
        self.function = function
        self._future: Future = Future()

    def run(self):
        if self.function is None:
            raise NotImplementedError
        return self.function()

    def start(self):
        QCoreApplication.postEvent(self, QEvent(Task.__ExecuteCall))

    def future(self) -> Future:
        return self._future

    def result(self, timeout: Optional[float] = None):
        return self._future.result(timeout)

    def _execute(self):
        try:
            if not self._future.set_running_or_notify_cancel():
                self.cancelled.emit()
                return
            self.started.emit()
            result = self.run()
            self._future.set_result(result)
            self.resultReady.emit(result)
        except BaseException as ex:
            self._future.set_exception(ex)
            self.exceptionReady.emit(ex)
        finally:
            self.finished.emit()

    def customEvent(self, event: QEvent):
        if event.type() == Task.__ExecuteCall:
            self._execute()
        else:
            super().customEvent(event)


class FutureWatcher(QObject):
    done = Signal(Future)
    finished = Signal(Future)
    cancelled = Signal(Future)
    resultReady = Signal(object)
    exceptionReady = Signal(BaseException)

    __FutureDone = QEvent.registerEventType()

    def __init__(self, future: Optional[Future] = None, parent: Optional[QObject] = None):
        super().__init__(parent)
        self.__future: Optional[Future] = None
        if future is not None:
            self.setFuture(future)

    def setFuture(self, future: Future):
        if self.__future is not None:
            raise RuntimeError("Future already set")
        self.__future = future

        def on_done(f):
            if (selfref := weakref.ref(self)) is None:
                return
            try:
                QCoreApplication.postEvent(selfref(), QEvent(FutureWatcher.__FutureDone))
            except RuntimeError:
                pass

        future.add_done_callback(on_done)

    def future(self) -> Future:
        return self.__future

    def __emitSignals(self):
        f = self.__future
        if f.cancelled():
            self.cancelled.emit(f)
            self.done.emit(f)
        else:
            self.finished.emit(f)
            self.done.emit(f)
            if exc := f.exception():
                self.exceptionReady.emit(exc)
            else:
                self.resultReady.emit(f.result())

    def customEvent(self, event: QEvent):
        if event.type() == FutureWatcher.__FutureDone:
            self.__emitSignals()
        super().customEvent(event)


# --------------------------------------------------------------------------- #
# Modern thread_worker utilities (QRunnable-based)
# --------------------------------------------------------------------------- #
_T = TypeVar("_T")
_R = TypeVar("_R")
_Y = TypeVar("_Y")
_S = TypeVar("_S")
_P = ParamSpec("_P")


class WorkerBaseSignals(QObject):
    started = Signal()
    finished = Signal()
    _finished = Signal(object)
    returned = Signal(object)
    errored = Signal(object)
    warned = Signal(tuple)


class WorkerBase(QRunnable, Generic[_R]):
    _worker_set: set[Self] = set()

    def __init__(self, SignalsClass=WorkerBaseSignals):
        super().__init__()
        self._abort_requested = False
        self._running = False
        self.signals = SignalsClass()

    def __getattr__(self, name: str):
        attr = getattr(self.signals.__class__, name, None)
        if isinstance(attr, Signal):
            return getattr(self.signals, name)
        raise AttributeError(f"{self.__class__.__name__!r} has no attribute {name!r}")

    def quit(self):
        self._abort_requested = True

    @property
    def abort_requested(self) -> bool:
        return self._abort_requested

    @property
    def is_running(self) -> bool:
        return self._running

    def run(self):
        self.started.emit()
        self._running = True
        try:
            with warnings.catch_warnings():
                warnings.filterwarnings("always")
                warnings.showwarning = lambda *w: self.warned.emit(w)
                result = self.work()
            if isinstance(result, Exception):
                raise result
            if not self.abort_requested:
                self.returned.emit(result)
        except Exception as e:
            self.errored.emit(e)
        finally:
            self._running = False
            self.finished.emit()
            self._finished.emit(self)

    def work(self) -> _R | Exception:
        raise NotImplementedError

    def start(self):
        if self in self._worker_set:
            raise RuntimeError("Worker already started")
        self._worker_set.add(self)
        self._finished.connect(lambda w: self._worker_set.discard(w))
        QThreadPool.globalInstance().start(self)


class FunctionWorker(WorkerBase[_R]):
    def __init__(self, func: Callable[_P, _R], *args, **kwargs):
        if inspect.isgeneratorfunction(func):
            raise TypeError("Use GeneratorWorker for generator functions")
        super().__init__()
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def work(self) -> _R:
        return self._func(*self._args, **self._kwargs)


class GeneratorWorkerSignals(WorkerBaseSignals):
    yielded = Signal(object)
    paused = Signal()
    resumed = Signal()
    aborted = Signal()


class GeneratorWorker(WorkerBase, Generic[_Y, _S, _R]):
    yielded: Signal
    paused: Signal
    resumed: Signal
    aborted: Signal

    def __init__(self, func: Callable[..., Generator[_Y, Optional[_S], _R]], *args, **kwargs):
        if not inspect.isgeneratorfunction(func):
            raise TypeError("Use FunctionWorker for regular functions")
        super().__init__(SignalsClass=GeneratorWorkerSignals)
        self._gen = func(*args, **kwargs)
        self._incoming: Optional[_S] = None
        self._pause_requested = self._resume_requested = self._paused = False
        self._pause_interval = 0.01

    def work(self) -> _R | None:
        while True:
            if self.abort_requested:
                self.aborted.emit()
                break
            if self._paused:
                if self._resume_requested:
                    self._paused = self._resume_requested = False
                    self.resumed.emit()
                else:
                    time.sleep(self._pause_interval)
                    continue
            elif self._pause_requested:
                self._paused = True
                self._pause_requested = False
                self.paused.emit()
                continue
            try:
                output = self._gen.send(self._incoming)
                self.yielded.emit(output)
                self._incoming = None
            except StopIteration as e:
                return e.value
        return None

    def send(self, value: _S):
        self._incoming = value

    def toggle_pause(self):
        if self._paused:
            self._resume_requested = True
        else:
            self._pause_requested = True


@overload
def create_worker(
    func: Callable[_P, Generator[_Y, _S, _R]], *args, **kwargs
) -> GeneratorWorker[_Y, _S, _R]: ...
@overload
def create_worker(func: Callable[_P, _R], *args, **kwargs) -> FunctionWorker[_R]: ...

def create_worker(func, *args, **kwargs):
    WorkerCls = GeneratorWorker if inspect.isgeneratorfunction(func) else FunctionWorker
    return WorkerCls(func, *args, **kwargs)


def thread_worker(func=None, **dec_kwargs):
    def decorator(f):
        @wraps(f)
        def wrapper(*a, **kw):
            return create_worker(f, *a, **kw, **dec_kwargs)
        return wrapper
    return decorator if func is None else decorator(func)


# --------------------------------------------------------------------------- #
# QThread-based worker helper (alternative to QRunnable)
# --------------------------------------------------------------------------- #
def new_worker_qthread(
    Worker: type[QObject],
    *args,
    _start_thread: bool = False,
    _connect: Optional[Dict[str, Callable]] = None,
    **kwargs,
):
    thread = QThread()
    worker = Worker(*args, **kwargs)
    worker.moveToThread(thread)
    thread.started.connect(worker.work)
    worker.finished.connect(thread.quit)
    worker.finished.connect(worker.deleteLater)
    thread.finished.connect(thread.deleteLater)
    if _connect:
        for sig, slot in _connect.items():
            getattr(worker, sig).connect(slot)
    if _start_thread:
        thread.start()
    return worker, thread

# --------------------------------------------------------------------------- #
# ProcessPoolExecutor - multiprocessing-based executor with Qt signal support
# --------------------------------------------------------------------------- #
import multiprocessing
import multiprocessing.pool
import queue

# Use spawn context by default to avoid fork+Qt deadlocks on POSIX systems.
# fork() can cause deadlocks when Qt threads are active (Qt's thread pool holds
# locks that get copied to child process but the threads that would unlock them
# don't exist in the child). spawn is already default on Windows and macOS.
# See: https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
_default_mp_context = multiprocessing.get_context('spawn')


class ProcessPoolExecutorSignals(QObject):
    """Qt signals for ProcessPoolExecutor events."""
    task_submitted = Signal(object)  # future
    task_completed = Signal(object)  # future
    task_failed = Signal(object, object)  # future, exception
    pool_shutdown = Signal()


class ProcessPoolExecutor(QObject, concurrent.futures.Executor):
    """
    A ProcessPoolExecutor that provides concurrent.futures.ProcessPoolExecutor API
    with optional Qt signal integration for task completion notifications.

    Unlike ThreadExecutor which uses QThreadPool, this uses Python's multiprocessing
    for true parallel execution of CPU-bound tasks across multiple cores.

    Usage:
        executor = ProcessPoolExecutor(max_workers=4)

        # Standard concurrent.futures API
        future = executor.submit(cpu_bound_task, arg1, arg2)
        result = future.result()

        # Or with Qt signals
        executor.signals.task_completed.connect(on_task_done)
        future = executor.submit(cpu_bound_task, arg1, arg2)

        executor.shutdown(wait=True)

    Note: Functions submitted must be picklable (module-level functions, not lambdas
    or closures that capture unpicklable objects).
    """

    def __init__(
        self,
        max_workers: Optional[int] = None,
        mp_context: Optional[multiprocessing.context.BaseContext] = None,
        initializer: Optional[Callable[..., None]] = None,
        initargs: Tuple = (),
        parent: Optional[QObject] = None,
    ):
        super().__init__(parent)
        self.signals = ProcessPoolExecutorSignals()
        self._max_workers = max_workers or multiprocessing.cpu_count()
        # Use spawn context by default to avoid fork+Qt deadlocks
        self._mp_context = mp_context if mp_context is not None else _default_mp_context
        self._initializer = initializer
        self._initargs = initargs

        # Create the underlying ProcessPoolExecutor
        self._executor = concurrent.futures.ProcessPoolExecutor(
            max_workers=self._max_workers,
            mp_context=self._mp_context,
            initializer=self._initializer,
            initargs=self._initargs,
        )

        self._futures: List[concurrent.futures.Future] = []
        self._shutdown = False
        self._state_lock = threading.Lock()

    @property
    def max_workers(self) -> int:
        """Return the maximum number of worker processes."""
        return self._max_workers

    def submit(self, fn: Callable, *args, **kwargs) -> concurrent.futures.Future:
        """
        Submit a callable to be executed in a worker process.

        Args:
            fn: A picklable callable to execute
            *args: Positional arguments for the callable
            **kwargs: Keyword arguments for the callable

        Returns:
            A Future representing the pending execution

        Raises:
            RuntimeError: If the executor has been shut down
        """
        with self._state_lock:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown.")

            future = self._executor.submit(fn, *args, **kwargs)
            self._futures.append(future)

            # Add callback for Qt signal emission
            future.add_done_callback(self._on_future_done)

            # Emit task_submitted signal
            try:
                self.signals.task_submitted.emit(future)
            except RuntimeError:
                pass  # Qt object may have been deleted

            return future

    def _on_future_done(self, future: concurrent.futures.Future):
        """Called when a future completes - emits appropriate Qt signals."""
        with self._state_lock:
            if future in self._futures:
                self._futures.remove(future)

        try:
            exc = future.exception()
            if exc is not None:
                self.signals.task_failed.emit(future, exc)
            else:
                self.signals.task_completed.emit(future)
        except (concurrent.futures.CancelledError, RuntimeError):
            pass  # Future was cancelled or Qt object deleted

    def map(
        self,
        fn: Callable,
        *iterables,
        timeout: Optional[float] = None,
        chunksize: int = 1,
    ):
        """
        Map a function over iterables, executing in parallel.

        Args:
            fn: A picklable callable
            *iterables: Iterables of arguments
            timeout: Maximum time to wait for results
            chunksize: Size of chunks for efficiency (larger = fewer IPC calls)

        Returns:
            Iterator of results in the same order as the input iterables

        Raises:
            RuntimeError: If the executor has been shut down
        """
        with self._state_lock:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown.")

        return self._executor.map(fn, *iterables, timeout=timeout, chunksize=chunksize)

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False):
        """
        Shutdown the executor.

        Args:
            wait: If True, wait for all pending futures to complete
            cancel_futures: If True, cancel all pending futures
        """
        with self._state_lock:
            self._shutdown = True

        self._executor.shutdown(wait=wait, cancel_futures=cancel_futures)

        try:
            self.signals.pool_shutdown.emit()
        except RuntimeError:
            pass  # Qt object may have been deleted

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False


# Alias for consistency with naming convention
QProcessPoolExecutor = ProcessPoolExecutor
QThreadPoolExecutor = ThreadExecutor  # Alias


# --------------------------------------------------------------------------- #
# InterpreterPoolExecutor - process-based executor with Qt signal support
# Mimics Python 3.13+ InterpreterPoolExecutor API using ProcessPoolExecutor
# for embedded contexts (like IDA Pro) where subinterpreters won't work
# --------------------------------------------------------------------------- #

# Always available - uses ProcessPoolExecutor as backend for isolation
INTERPRETER_POOL_AVAILABLE = True


class InterpreterPoolExecutorSignals(QObject):
    """Qt signals for InterpreterPoolExecutor events."""
    task_submitted = Signal(object)  # future
    task_completed = Signal(object)  # future
    task_failed = Signal(object, object)  # future, exception
    pool_shutdown = Signal()


class InterpreterPoolExecutor(QObject, concurrent.futures.Executor):
    """
    An executor that provides concurrent.futures.InterpreterPoolExecutor API
    with optional Qt signal integration for task completion notifications.

    This implementation uses ProcessPoolExecutor as the backend to provide
    true parallelism and isolation, making it compatible with embedded Python
    contexts (like IDA Pro) where Python 3.13+ subinterpreters may not work.

    Usage:
        executor = InterpreterPoolExecutor(max_workers=4)

        # Standard concurrent.futures API
        future = executor.submit(cpu_bound_task, arg1, arg2)
        result = future.result()

        # Or with Qt signals
        executor.signals.task_completed.connect(on_task_done)
        future = executor.submit(cpu_bound_task, arg1, arg2)

        executor.shutdown(wait=True)

    Note: Functions and arguments must be picklable (same restrictions as
    ProcessPoolExecutor/multiprocessing).
    """

    def __init__(
        self,
        max_workers: Optional[int] = None,
        parent: Optional[QObject] = None,
    ):
        super().__init__(parent)
        self.signals = InterpreterPoolExecutorSignals()
        self._max_workers = max_workers or multiprocessing.cpu_count()

        # Use ProcessPoolExecutor as backend for isolation
        # This provides similar benefits to subinterpreters (no GIL contention)
        # while being compatible with embedded Python contexts
        # Use spawn context to avoid fork+Qt deadlocks
        self._executor = concurrent.futures.ProcessPoolExecutor(
            max_workers=self._max_workers,
            mp_context=_default_mp_context,
        )

        self._futures: List[concurrent.futures.Future] = []
        self._shutdown = False
        self._state_lock = threading.Lock()

    @property
    def max_workers(self) -> int:
        """Return the maximum number of worker processes."""
        return self._max_workers

    def submit(self, fn: Callable, *args, **kwargs) -> concurrent.futures.Future:
        """
        Submit a callable to be executed in a worker process.

        Args:
            fn: A callable to execute (must be picklable)
            *args: Positional arguments for the callable
            **kwargs: Keyword arguments for the callable

        Returns:
            A Future representing the pending execution

        Raises:
            RuntimeError: If the executor has been shut down
        """
        with self._state_lock:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown.")

            future = self._executor.submit(fn, *args, **kwargs)
            self._futures.append(future)

            # Add callback for Qt signal emission
            future.add_done_callback(self._on_future_done)

            # Emit task_submitted signal
            try:
                self.signals.task_submitted.emit(future)
            except RuntimeError:
                pass  # Qt object may have been deleted

            return future

    def _on_future_done(self, future: concurrent.futures.Future):
        """Called when a future completes - emits appropriate Qt signals."""
        with self._state_lock:
            if future in self._futures:
                self._futures.remove(future)

        try:
            exc = future.exception()
            if exc is not None:
                self.signals.task_failed.emit(future, exc)
            else:
                self.signals.task_completed.emit(future)
        except (concurrent.futures.CancelledError, RuntimeError):
            pass  # Future was cancelled or Qt object deleted

    def map(
        self,
        fn: Callable,
        *iterables,
        timeout: Optional[float] = None,
        chunksize: int = 1,
    ):
        """
        Map a function over iterables, executing in parallel processes.

        Args:
            fn: A callable (must be picklable)
            *iterables: Iterables of arguments
            timeout: Maximum time to wait for results
            chunksize: Size of chunks for efficiency

        Returns:
            Iterator of results in the same order as the input iterables

        Raises:
            RuntimeError: If the executor has been shut down
        """
        with self._state_lock:
            if self._shutdown:
                raise RuntimeError("Cannot schedule new futures after shutdown.")

        return self._executor.map(fn, *iterables, timeout=timeout, chunksize=chunksize)

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False):
        """
        Shutdown the executor.

        Args:
            wait: If True, wait for all pending futures to complete
            cancel_futures: If True, cancel all pending futures
        """
        with self._state_lock:
            self._shutdown = True

        self._executor.shutdown(wait=wait, cancel_futures=cancel_futures)

        try:
            self.signals.pool_shutdown.emit()
        except RuntimeError:
            pass  # Qt object may have been deleted

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=True)
        return False


# Alias for naming consistency
QInterpreterPoolExecutor = InterpreterPoolExecutor


# Single-file ready-to-use module.
# No external dependencies beyond PyQt5/PySide6.

__all__ = [
    # Asyncio integration
    "QAsyncioEventLoop",
    "QAsyncioEventLoopPolicy",
    "run",
    "set_event_loop_policy",
    # Thread executor
    "ThreadExecutor",
    "QThreadPoolExecutor",  # Alias
    "ThreadPoolExecutorSignals",
    "Task",
    "FutureWatcher",
    "Future",
    # Process executor
    "ProcessPoolExecutor",
    "QProcessPoolExecutor",  # Alias
    "ProcessPoolExecutorSignals",
    # Interpreter executor (Python 3.13+)
    "InterpreterPoolExecutor",
    "QInterpreterPoolExecutor",  # Alias
    "InterpreterPoolExecutorSignals",
    "INTERPRETER_POOL_AVAILABLE",
    # Worker utilities
    "WorkerBase",
    "WorkerBaseSignals",
    "FunctionWorker",
    "GeneratorWorker",
    "GeneratorWorkerSignals",
    "create_worker",
    "thread_worker",
    "new_worker_qthread",
]



import sys
sys.path.insert(0, "/home/user/ida-taskr/src")

with open("/tmp/ida_imports_test.txt", "w") as f:
    f.write("Testing imports...\n\n")

    try:
        f.write("1. Importing ida_taskr...\n")
        import ida_taskr
        f.write("   [OK] Success\n\n")

        f.write("2. Checking is_ida()...\n")
        f.write(f"   Result: {ida_taskr.is_ida()}\n\n")

        f.write("3. Importing MessageEmitter...\n")
        from ida_taskr import MessageEmitter
        f.write("   [OK] Success\n\n")

        f.write("4. Creating MessageEmitter...\n")
        emitter = MessageEmitter()
        f.write(f"   [OK] Created: {emitter}\n\n")

        f.write("5. Importing WorkerLauncher...\n")
        from ida_taskr import WorkerLauncher
        f.write("   [OK] Success\n\n")

        f.write("6. Creating WorkerLauncher...\n")
        launcher = WorkerLauncher(emitter)
        f.write(f"   [OK] Created: {launcher}\n\n")

        f.write("ALL TESTS PASSED!\n")

    except Exception as e:
        f.write(f"\nError: {e}\n")
        import traceback
        f.write(traceback.format_exc())

try:
    import idaapi
    idaapi.qexit(0)
except:
    pass



"""
TaskRunner - A simplified interface for running worker tasks.

Provides a high-level API that wraps WorkerLauncher and MessageEmitter
for easier task execution with callback-based result handling.
"""

import logging

from .helpers import get_logger
from .launcher import WorkerLauncher
from .protocols import MessageEmitter


class TaskRunner:
    """Simplified task runner with callback-based event handling."""

    def __init__(self, worker_script, worker_args, log_level=None, logger=None):
        """Initialize TaskRunner.

        Args:
            worker_script: Path to the worker script to execute
            worker_args: Dictionary of arguments to pass to the worker
            log_level: Logging level (optional)
            logger: Custom logger instance (optional)
        """
        if logger:
            self.logger = logger
        else:
            # Use INFO as default if no log_level specified
            actual_log_level = log_level if log_level is not None else logging.INFO
            self.logger = get_logger(log_level=actual_log_level)
        self.message_emitter = MessageEmitter()
        self.launcher = WorkerLauncher(self.message_emitter)
        self.worker_script = worker_script
        self.worker_args = worker_args
        self._results_callback = None
        self._progress_callback = None

    def on_results(self, callback):
        """Register a callback for when worker results are received.

        Args:
            callback: Function to call with results dict
        """
        self._results_callback = callback
        self.message_emitter.on("worker_results", self._handle_results)

    def on_progress(self, callback):
        """Register a callback for progress updates.

        Args:
            callback: Function to call with (progress, status) tuple
        """
        self._progress_callback = callback
        self.message_emitter.on("worker_message", self._handle_progress)

    def start(self):
        """Start the worker task."""
        if self.launcher.launch_worker(self.worker_script, self.worker_args):
            self.logger.info("Worker launched successfully")
        else:
            self.logger.error("Failed to launch worker")

    def _handle_results(self, results):
        """Internal handler for worker results."""
        if self._results_callback:
            self._results_callback(results)

    def _handle_progress(self, message):
        """Internal handler for progress messages."""
        if message.get("type") == "progress" and self._progress_callback:
            progress = message.get("progress", 0)
            status = message.get("status", "unknown")
            self._progress_callback(progress, status)



"""
IDA Worker Manager - A reusable task worker manager library for IDA Pro.

This library provides a Qt-based multiprocessing framework for running
CPU-intensive tasks outside of IDA's main thread while maintaining
bidirectional communication.
"""

from .helpers import MultiprocessingHelper, get_logger, is_ida
from .launcher import WorkerLauncher
from .protocols import MessageEmitter, WorkerProtocol

# New import for TaskRunner
from .task_runner import TaskRunner
from .utils import DataProcessorCore
from .worker import ConnectionContext, WorkerBase

# Decorators for simplified API
from .decorators import background_task, cpu_task, io_task, parallel, shared_memory_task

# QtAsyncio integration (optional)
from .qt_compat import QT_ASYNCIO_AVAILABLE

# Always define INTERPRETER_POOL_AVAILABLE (False if Qt not available)
INTERPRETER_POOL_AVAILABLE = False

__all__ = [
    "WorkerLauncher",
    "WorkerBase",
    "ConnectionContext",
    "MultiprocessingHelper",
    "get_logger",
    "is_ida",
    "WorkerProtocol",
    "MessageEmitter",
    "TaskRunner",
    "DataProcessorCore",
    "QT_ASYNCIO_AVAILABLE",
    "INTERPRETER_POOL_AVAILABLE",
    # Decorators
    "background_task",
    "cpu_task",
    "io_task",
    "parallel",
    "shared_memory_task",
]

# Conditionally export qtasyncio utilities if available
if QT_ASYNCIO_AVAILABLE:
    try:
        from .qtasyncio import (
            # Asyncio integration
            QAsyncioEventLoop,
            QAsyncioEventLoopPolicy,
            run as qtasyncio_run,
            set_event_loop_policy,
            # Thread executor
            ThreadExecutor,
            QThreadPoolExecutor,
            Task,
            FutureWatcher,
            # Process executor
            ProcessPoolExecutor,
            QProcessPoolExecutor,
            # Interpreter executor
            InterpreterPoolExecutor,
            QInterpreterPoolExecutor,
            INTERPRETER_POOL_AVAILABLE as _INTERPRETER_POOL_AVAILABLE,
            # Worker utilities
            WorkerBase as QtWorkerBase,
            FunctionWorker,
            GeneratorWorker,
            create_worker,
            thread_worker,
            new_worker_qthread,
        )

        # Update module-level variable with actual value from qtasyncio
        INTERPRETER_POOL_AVAILABLE = _INTERPRETER_POOL_AVAILABLE

        __all__.extend([
            # Asyncio integration
            "QAsyncioEventLoop",
            "QAsyncioEventLoopPolicy",
            "qtasyncio_run",
            "set_event_loop_policy",
            # Thread executor
            "ThreadExecutor",
            "QThreadPoolExecutor",
            "Task",
            "FutureWatcher",
            # Process executor
            "ProcessPoolExecutor",
            "QProcessPoolExecutor",
            # Interpreter executor (Python 3.13+)
            "InterpreterPoolExecutor",
            "QInterpreterPoolExecutor",
            "INTERPRETER_POOL_AVAILABLE",
            # Worker utilities
            "QtWorkerBase",
            "FunctionWorker",
            "GeneratorWorker",
            "create_worker",
            "thread_worker",
            "new_worker_qthread",
        ])
    except ImportError:
        # QtAsyncio module not available
        pass

__version__ = "1.0.0"



"""Protocol definitions and base classes for the worker manager."""

from abc import ABC, abstractmethod

from .utils import EventEmitter


class WorkerProtocol(ABC):
    """Protocol for worker implementations."""

    @abstractmethod
    def setup(self, **kwargs):
        """Set up the worker."""
        pass

    @abstractmethod
    def process(self, connection, **kwargs):
        """Process tasks."""
        pass

    @abstractmethod
    def cleanup(self):
        """Clean up resources."""
        pass


class MessageEmitter(EventEmitter):
    """Event emitter for handling messages from workers in IDA.

    Events emitted:
    - 'worker_connected': When worker establishes connection
    - 'worker_message': When a message is received from worker (payload: message dict)
    - 'worker_results': When results are received from worker (payload: results dict)
    - 'worker_error': When an error occurs (payload: error string)
    - 'worker_disconnected': When worker connection is closed
    """

    def emit_worker_connected(self):
        """Emit worker connected event."""
        self.emit("worker_connected")

    def emit_worker_message(self, message: dict):
        """Emit worker message event."""
        self.emit("worker_message", message)

    def emit_worker_results(self, results: dict):
        """Emit worker results event."""
        self.emit("worker_results", results)

    def emit_worker_error(self, error: str):
        """Emit worker error event."""
        self.emit("worker_error", error)

    def emit_worker_disconnected(self):
        """Emit worker disconnected event."""
        self.emit("worker_disconnected")



"""Test if QProcess works in headless IDA."""

import sys
sys.path.insert(0, "/home/user/ida-taskr/src")

with open("/tmp/ida_qprocess_test.txt", "w") as f:
    f.write("Testing QProcess in headless IDA\n")
    f.write("=" * 70 + "\n\n")

    try:
        import idaapi
        f.write(f"IDA version: {idaapi.get_kernel_version()}\n\n")

        from PySide6.QtCore import QCoreApplication, QProcess
        app = QCoreApplication.instance()
        f.write(f"Qt app: {type(app).__name__}\n")
        f.write(f"Qt app exists: {app is not None}\n\n")

        # Test 1: Create a QProcess
        f.write("Test 1: Creating QProcess...\n")
        process = QProcess()
        f.write(f"[OK] QProcess created: {process}\n\n")

        # Test 2: Run a simple command
        f.write("Test 2: Running 'echo hello' command...\n")
        process.start("echo", ["hello"])

        # Wait for it to finish
        if process.waitForFinished(1000):  # 1 second timeout
            output = process.readAllStandardOutput().data().decode()
            f.write(f"[OK] Command executed\n")
            f.write(f"     Output: {output.strip()}\n\n")
        else:
            f.write(f"[WARN] Command timed out or failed\n")
            f.write(f"       State: {process.state()}\n\n")

        # Test 3: Create WorkerLauncher (uses QProcess)
        f.write("Test 3: Creating WorkerLauncher...\n")
        from ida_taskr import WorkerLauncher, MessageEmitter

        emitter = MessageEmitter()
        launcher = WorkerLauncher(emitter)
        f.write(f"[OK] WorkerLauncher created: {launcher}\n\n")

        # Test 4: Verify WorkerLauncher has QProcess methods
        f.write("Test 4: Verifying WorkerLauncher QProcess interface...\n")
        f.write(f"     Has start(): {hasattr(launcher, 'start')}\n")
        f.write(f"     Has waitForFinished(): {hasattr(launcher, 'waitForFinished')}\n")
        f.write(f"     Has readAllStandardOutput(): {hasattr(launcher, 'readAllStandardOutput')}\n")
        f.write(f"     Inherits QProcess: {isinstance(launcher, QProcess)}\n\n")

        f.write("=" * 70 + "\n")
        f.write("ALL TESTS PASSED - QProcess works in headless IDA!\n")
        f.write("=" * 70 + "\n")

    except Exception as e:
        f.write(f"\n[FAIL] Error: {e}\n")
        import traceback
        f.write(traceback.format_exc())
        f.write("=" * 70 + "\n")

try:
    import idaapi
    idaapi.qexit(0)
except:
    pass



"""IDA-side components for launching and managing worker processes."""

import multiprocessing
import multiprocessing.connection
import os
import pickle
import select
import sys

from .helpers import MultiprocessingHelper, get_logger
from .protocols import MessageEmitter
from .qt_compat import QtCore, Signal, QProcessEnvironment, QT_AVAILABLE

logger = get_logger()


class TemporarilyDisableNotifier:
    """Context manager to temporarily disable a QSocketNotifier."""

    def __init__(self, notifier):
        self.notifier = notifier
        self.was_enabled = False

    def __enter__(self):
        self.was_enabled = self.notifier.isEnabled()
        self.notifier.setEnabled(False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.was_enabled:
            self.notifier.setEnabled(True)


class ConnectionReader(QtCore.QThread):
    """Qt thread for reading messages from worker connection."""

    message_received = Signal(object)
    connection_closed = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.connection: multiprocessing.connection.Connection | None = None

    def set_connection(self, conn):
        self.connection = conn

    def run(self):
        if not self.connection:
            return

        try:
            while True:
                msg = self.connection.recv()
                self.message_received.emit(msg)
        except (EOFError, OSError):
            self.connection_closed.emit()
        except pickle.PickleError as e:
            logger.error(f"Pickle error: {e}")
        finally:
            try:
                self.connection.close()
            except Exception:
                pass


class QtListener(QtCore.QObject):
    """Qt-friendly wrapper for multiprocessing.connection.Listener."""

    family = "AF_INET"
    connection_accepted = Signal(object)
    connection_error = Signal(str)

    def __init__(self, address=None, backlog=1, authkey=None, parent=None):
        super().__init__(parent)

        self._listener = multiprocessing.connection.Listener(
            address, self.family, backlog, authkey
        )

        self._socket = self._listener._listener._socket  # type: ignore
        self._notifier = QtCore.QSocketNotifier(
            self._socket.fileno(), QtCore.QSocketNotifier.Read, self
        )
        self._notifier.activated.connect(self._on_connection_ready)
        self._notifier.setEnabled(True)

    def _on_connection_ready(self):
        with TemporarilyDisableNotifier(self._notifier):
            try:
                ready, _, _ = select.select([self._socket], [], [], 0)
                if not ready:
                    return

                conn = self.accept()
                self.connection_accepted.emit(conn)
            except Exception as e:
                logger.error(f"Error accepting connection: {e}")
                self.connection_error.emit(str(e))

    def accept(self):
        return self._listener.accept()

    def close(self):
        self._notifier.setEnabled(False)
        self._notifier.deleteLater()
        self._listener.close()

    @property
    def address(self):
        return self._listener.address


class WorkerLauncher(QtCore.QProcess):
    """
    Manages external worker processes using QProcess.
    Provides bidirectional communication via multiprocessing connections.
    """

    # Signals
    processing_results = Signal(dict)
    error_occurred_msg = Signal(str)
    worker_message = Signal(object)
    worker_connected = Signal()
    worker_disconnected = Signal()

    def __init__(self, message_emitter: MessageEmitter | None = None, parent=None):
        super(WorkerLauncher, self).__init__(parent)

        self.message_emitter = message_emitter

        self.readyReadStandardOutput.connect(self._on_stdout)
        self.readyReadStandardError.connect(self._on_stderr)
        self.errorOccurred.connect(self._on_error)
        self.stateChanged.connect(self._on_state_changed)

        self.python_interpreter = MultiprocessingHelper.get_python_interpreter()

        self.listener = None
        self.connection = None
        self.authkey = None

        self.reader_thread = ConnectionReader(self)
        self.reader_thread.message_received.connect(self._on_worker_message)
        self.reader_thread.connection_closed.connect(self._on_connection_closed)

        self.connection_attempts = 0
        self.max_connection_attempts = 10

        self._streams: dict[str, dict] = {}

    def is_not_running(self):
        return self.state() == QtCore.QProcess.NotRunning

    def _on_worker_message(self, message):
        """Process messages from the worker, including chunked streams."""
        # Handle chunked messages
        msg_id = message.get("message_id")
        if msg_id:
            idx = message["chunk_index"]
            total = message["total_chunks"]
            msg_type = message.get("type")

            stream = self._streams.setdefault(
                msg_id, {"type": msg_type, "chunks": {}, "total": total}
            )
            stream["chunks"][idx] = message["data"]
            logger.info(
                "Received chunk %d/%d for %r (id=%s)",
                idx + 1,
                total,
                msg_type,
                msg_id,
            )

            if len(stream["chunks"]) == total:
                full = []
                for i in range(total):
                    full.extend(stream["chunks"][i])
                del self._streams[msg_id]
                logger.info(
                    "%r streaming complete (id=%s, %d items)",
                    msg_type,
                    msg_id,
                    len(full),
                )
                self.processing_results.emit(
                    {
                        "type": "result",
                        "results": full,
                        "status": "success",
                    }
                )
            return

        # Non-chunked messages
        logger.debug("← Received message from worker: %r", message)
        self.worker_message.emit(message)

        # Delegate to MessageEmitter if available
        if self.message_emitter and isinstance(message, dict):
            msg_type = message.get("type")
            if msg_type == "error":
                self.message_emitter.emit_worker_error(
                    message.get("error", "Unknown error")
                )
            elif msg_type == "result":
                self.message_emitter.emit_worker_results(
                    {
                        "results": message.get("data"),
                        "status": message.get("status", "success"),
                    }
                )
            elif msg_type == "status":
                self.message_emitter.emit_worker_message(message)

    def _on_connection_closed(self):
        logger.info("Worker IPC connection closed.")
        if self.connection:
            try:
                self.connection.close()
            except:
                pass
            self.connection = None
        self.worker_disconnected.emit()

        if self.message_emitter:
            self.message_emitter.emit_worker_disconnected()

    def launch_worker(self, script_path: str, worker_args: dict):
        """
        Starts the worker script with specified arguments.

        Args:
            script_path: Path to the worker script
            worker_args: Dictionary of arguments to pass to the worker
        """
        self._cleanup_resources()

        # Create listener
        self.authkey = os.urandom(32)
        logger.info(f"Generated Authkey: {self.authkey.hex()}")
        self.listener = QtListener(
            ("localhost", 0),
            authkey=self.authkey,
            parent=self,
        )
        address = self.listener.address
        logger.info(f"Created listener on {address}")
        self.listener.connection_accepted.connect(self._on_connection_accepted)
        self.listener.connection_error.connect(self._on_connection_error)

        # Prepare QProcess
        env = QProcessEnvironment.systemEnvironment()
        env.insert("PYTHON_PATH", str(self.python_interpreter.parent))
        env.insert("PYTHON_BIN", str(self.python_interpreter.name))
        self.setProcessEnvironment(env)

        # Build command line arguments
        args = ["-u", script_path]

        # Add connection info
        args.extend(
            [
                "--address",
                f"{address[0]}:{address[1]}",
                "--authkey",
                self.authkey.hex(),
            ]
        )

        # Add worker-specific arguments
        for key, value in worker_args.items():
            args.extend([f"--{key}", str(value)])

        logger.info(f"Starting worker process: {self.python_interpreter} {args}")
        self.start(str(self.python_interpreter), args)

        if not self.waitForStarted(5000):
            logger.error(f"Worker process failed to start: {self.errorString()}")
            self._cleanup_resources()
            return False

        self.connection_attempts = 0
        logger.info("Worker process started. Beginning connection attempts...")
        return True

    def _on_connection_accepted(self, conn):
        logger.info("Connection from worker accepted")
        self.reader_thread.set_connection(conn)
        if not self.reader_thread.isRunning():
            self.reader_thread.start()
        self.connection = conn
        self.worker_connected.emit()

        if self.message_emitter:
            self.message_emitter.emit_worker_connected()

    def _on_connection_error(self, error_msg):
        logger.error(f"Connection error: {error_msg}")
        self.connection_attempts += 1

        if self.connection_attempts >= self.max_connection_attempts:
            self.error_occurred_msg.emit(
                f"Failed to connect to worker after {self.max_connection_attempts} attempts"
            )
            self.stop_worker()

    def _cleanup_resources(self):
        if self.reader_thread.isRunning():
            self.reader_thread.terminate()
            self.reader_thread.wait()

        if self.connection:
            try:
                self.connection.close()
            except:
                pass
            self.connection = None

        if self.listener:
            try:
                self.listener.close()
            except:
                pass
            self.listener = None

    def stop_worker(self):
        """Attempts to terminate the worker process gracefully."""
        if self.is_not_running() and not self.connection:
            logger.debug("Worker process was already stopped and connection closed.")
            return

        logger.info("Attempting to stop worker process...")

        if self.connection:
            try:
                logger.info("Sending exit command...")
                self.send_command({"command": "exit"})
                if self.waitForFinished(1000):
                    logger.info("Worker exited gracefully.")
                    self._cleanup_resources()
                    return
            except Exception as e:
                logger.error(f"Error sending exit command: {e}")

        if not self.is_not_running():
            logger.warning("Worker did not exit gracefully, terminating...")
            self.terminate()
            if not self.waitForFinished(2000):
                logger.warning("Worker did not terminate, killing...")
                self.kill()
                self.waitForFinished(1000)

        self._cleanup_resources()
        logger.info("Worker process shutdown complete.")

    def send_command(self, command):
        """Sends a command object via the connection."""
        if not self.connection:
            logger.warning(
                f"Cannot send command '{command}', IPC connection not established."
            )
            return False

        logger.debug(f"→ Sending command: {command}")
        try:
            self.connection.send(command)
            logger.debug(f"→ Successfully sent command: {command}")
            return True
        except Exception as e:
            logger.error(f"Failed to send command '{command}': {e}")
            self._on_connection_closed()
            return False

    def _on_stdout(self):
        out_bytes = self.readAllStandardOutput()
        if not out_bytes:
            return
        out = out_bytes.data().decode("utf-8", errors="replace")
        if out.strip():
            print(out.strip(), flush=True)

    def _on_stderr(self):
        data = self.readAllStandardError().data()
        if data:
            data = data.decode("utf-8", errors="replace").strip()
            print(data.strip(), file=sys.stderr, flush=True)

    def _on_error(self, error: QtCore.QProcess.ProcessError):
        error_map = {
            QtCore.QProcess.FailedToStart: "FailedToStart",
            QtCore.QProcess.Crashed: "Crashed",
            QtCore.QProcess.Timedout: "Timedout",
            QtCore.QProcess.ReadError: "ReadError",
            QtCore.QProcess.WriteError: "WriteError",
            QtCore.QProcess.UnknownError: "UnknownError",
        }
        error_str = error_map.get(error, f"UnknownError({error})")
        msg = f"Worker process error: {error_str} - {self.errorString()}"
        print(msg.strip(), file=sys.stderr, flush=True)
        self.error_occurred_msg.emit(msg)

    def _on_state_changed(self, state: QtCore.QProcess.ProcessState):
        state_map = {
            QtCore.QProcess.NotRunning: "NotRunning",
            QtCore.QProcess.Starting: "Starting",
            QtCore.QProcess.Running: "Running",
        }
        state_str = state_map.get(state, f"UnknownState({state})")
        logger.info(f"Worker process state changed: {state_str}")

        if self.is_not_running():
            self._cleanup_resources()
            exit_code = self.exitCode()
            exit_status = self.exitStatus()
            exit_status_str = (
                "NormalExit"
                if exit_status == QtCore.QProcess.NormalExit
                else "CrashExit"
            )
            msg = f"Worker process exited with code {exit_code} ({exit_status_str})"
            logger.info(msg)
            if exit_status == QtCore.QProcess.CrashExit:
                self.error_occurred_msg.emit(msg)



"""Utility functions for the anti-deobfuscation plugin."""

import asyncio
import atexit
import collections
import contextlib
import dataclasses
import enum
import functools
import logging
import multiprocessing
import multiprocessing.shared_memory
import pathlib
import time
import typing
from abc import ABC, abstractmethod
from bisect import bisect_left, bisect_right
from typing import Any, Callable, Generic, TypeVar, overload

from .helpers import get_logger

logger = get_logger(__name__)


def humanize_bytes(
    num_bytes: int, precision: int = 2, units: list[str] = ["B", "KB", "MB", "GB"]
) -> str:
    """
    Convert a byte count into a human-friendly string with units.
    """
    if num_bytes < 0:
        raise ValueError("num_bytes must be non-negative")
    if num_bytes == 0:
        return "0 B"
    idx = 0
    value = float(num_bytes)
    while value >= 1024 and idx < len(units) - 1:
        value /= 1024
        idx += 1
    if value.is_integer():
        return f"{int(value)} {units[idx]}"
    else:
        return f"{value:.{precision}f} {units[idx]}"


class DataProcessorCore:
    """Core processor for managing deobfuscation tasks."""

    _shared_memory = None

    def __init__(self, message_emitter):
        """Initialize the DataProcessorCore.

        Args:
            message_emitter: MessageEmitter instance to handle worker communication
        """
        from .protocols import MessageEmitter

        if not isinstance(message_emitter, MessageEmitter):
            raise TypeError("message_emitter must be a MessageEmitter instance")

        self.message_emitter = message_emitter
        self.proc = None
        atexit.register(self.terminate)

    @staticmethod
    def get_section_data(
        section_name: str,
        max_size: int = 120 * 1024 * 1024,
        min_size: int = 1024,
    ) -> tuple[int, bytes]:
        """Get the data of a section by name."""
        try:
            import ida_bytes
            import ida_segment
            import idaapi
        except ImportError:
            logger.error("IDA Pro modules not available")
            return 0, b""  # Use 0 instead of idaapi.BADADDR when not available

        seg = ida_segment.get_segm_by_name(section_name)
        if not seg:
            logger.error("Section %s not found", section_name)
            return idaapi.BADADDR, b""

        data_ea = seg.start_ea
        data_size = seg.end_ea - seg.start_ea

        if data_size > max_size:
            data_size = max_size
            logger.warning(
                "Limiting section data size to %s", humanize_bytes(data_size)
            )
        elif data_size < min_size:
            logger.error(
                "%s section is too small (%s)", section_name, humanize_bytes(data_size)
            )
            return idaapi.BADADDR, b""

        logger.info(
            "Reading %s from address %s", humanize_bytes(data_size), hex(data_ea)
        )

        data_bytes = ida_bytes.get_bytes(data_ea, data_size)
        if not data_bytes or len(data_bytes) != data_size:
            logger.error("Failed to read section data")
            return idaapi.BADADDR, b""

        return data_ea, data_bytes

    @staticmethod
    def from_range(start_ea: int, end_ea: int):
        """Get the data of a section by name and return the start address and the bytes."""
        try:
            import ida_bytes
        except ImportError:
            logger.error("IDA Pro modules not available")
            return 0, b""

        data_bytes = ida_bytes.get_bytes(start_ea, end_ea - start_ea)
        return start_ea, data_bytes

    def run(
        self, start_ea: int, bytes_to_process: bytes, worker_script_path: str, **kwargs
    ):
        """Run the deobfuscation process.

        Args:
            start_ea: Starting address for processing
            bytes_to_process: Binary data to process
            worker_script_path: Path to the worker script
            **kwargs: Additional arguments passed to worker
        """
        from .launcher import WorkerLauncher

        data_size = len(bytes_to_process)

        # Create shared memory
        self._shared_memory = multiprocessing.shared_memory.SharedMemory(
            create=True, size=data_size
        )
        self._shared_memory.buf[:data_size] = bytes_to_process

        # Launch worker with provided message emitter
        self.proc = WorkerLauncher(self.message_emitter)

        worker_args = {
            "shm_name": self._shared_memory.name,
            "data_size": data_size,
            "start_ea": hex(start_ea),
            **kwargs,
        }

        # Add IDA bitness if available
        try:
            import ida_ida

            worker_args["is64"] = "1" if ida_ida.inf_is_64bit() else "0"
        except ImportError:
            logger.warning("IDA Pro modules not available, bitness detection skipped")

        if not self.proc.launch_worker(str(worker_script_path), worker_args):
            self.terminate()
            logger.error("Failed to start worker process")
            return

    def terminate(self):
        """Terminate and clean up."""
        logger.info("Terminating...")
        if self.proc and not self.proc.is_not_running():
            self.proc.stop_worker()
            self.proc = None
        self._cleanup_shared_memory()
        logger.info("Terminated.")

    def _cleanup_shared_memory(self):
        """Clean up shared memory."""
        if not self._shared_memory:
            return
        try:
            self._shared_memory.close()
            shm = multiprocessing.shared_memory.SharedMemory(self._shared_memory.name)
            shm.unlink()
            logger.info("Shared memory unlinked: %s", self._shared_memory.name)
        except FileNotFoundError:
            logger.warning(
                "Shared memory already unlinked: %s", self._shared_memory.name
            )
        except PermissionError as e:
            logger.error("Permission error unlinking shared memory: %s", e)
        except Exception as e:
            logger.error(
                "Unexpected error unlinking shared memory: %s", e, exc_info=True
            )
        finally:
            self._shared_memory = None


class emit:
    def __init__(self, event):
        self.event = event

    def __call__(self, fn):
        @functools.wraps(fn)
        def wrapper(inst, *args, **kwargs):
            result = fn(inst, *args, **kwargs)
            inst.emit(self.event)
            return result

        return wrapper


T = TypeVar("T")


class reify(Generic[T]):
    """
    Acts similar to a property, except the result will be
    set as an attribute on the instance instead of recomputed
    each access.
    """

    def __init__(self, fn: Callable[..., T]) -> None:
        self.fn = fn
        # Copy function attributes to preserve metadata
        self.__name__ = getattr(fn, "__name__", "<unknown>")
        self.__doc__ = getattr(fn, "__doc__", None)
        self.__module__ = getattr(fn, "__module__", "") or ""
        self.__qualname__ = getattr(fn, "__qualname__", "") or ""
        self.__annotations__ = getattr(fn, "__annotations__", {})

    @overload
    def __get__(self, instance: None, owner: type) -> "reify[T]": ...

    @overload
    def __get__(self, instance: Any, owner: type) -> T: ...

    def __get__(self, instance: Any, owner: type) -> "T | reify[T]":
        if instance is None:
            return self

        fn = self.fn
        val = fn(instance)
        setattr(instance, fn.__name__, val)
        return val


class EventEmitter:
    @reify
    def _listeners(self):
        return collections.defaultdict(set)

    def on(self, event, handler=None):
        """Register an event handler for the given event."""
        if handler:
            self._listeners[event].add(handler)
            return handler

        @functools.wraps(self.on)
        def decorator(func):
            self.on(event, func)
            return func

        return decorator

    def once(self, event, handler):
        @functools.wraps(handler)
        def once_handler(*args, **kwargs):
            self.remove(event, once_handler)
            return handler(*args, **kwargs)

        self.on(event, once_handler)

    def remove(self, event, handler):
        self._listeners[event].discard(handler)

    def emit(self, event, *args, **kwargs):
        for handler in list(self._listeners[event]):
            handler(*args, **kwargs)


@dataclasses.dataclass
class AsyncEventEmitter(ABC):
    def __post_init__(self):
        self._listeners = collections.defaultdict(set)
        self.pause_evt = asyncio.Event()
        self.stop_evt = asyncio.Event()
        self.logger = get_logger(self.__class__.__name__)

    def on(self, event, handler=None):
        """Register an event handler for the given event."""
        if handler:
            self._listeners[event].add(handler)
            return handler

        @functools.wraps(self.on)
        def decorator(func):
            self.on(event, func)
            return func

        return decorator

    def once(self, event, handler):
        @functools.wraps(handler)
        def once_handler(*args, **kwargs):
            self.remove(event, once_handler)
            return handler(*args, **kwargs)

        self.on(event, once_handler)

    def remove(self, event, handler):
        self._listeners[event].discard(handler)

    async def emit(self, event, *args):
        for handler in list(self._listeners[event]):
            await handler(*args)

    @abstractmethod
    async def run(self):
        """Core asynchronous task execution logic."""
        pass

    @abstractmethod
    async def shutdown(self):
        """Cleanup resources for the async task."""
        pass


def log_execution_time(func, loglvl=logging.INFO):
    """
    Decorator to log the execution time of async stage methods.

    >>> import asyncio, logging
    >>> logging.basicConfig(level=logging.INFO)
    >>> class Dummy:
    ...     @log_execution_time
    ...     async def foo(self):
    ...         await asyncio.sleep(0.01)
    ...         return 42
    >>> d = Dummy()
    >>> asyncio.run(d.foo())
    42
    """

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = await func(*args, **kwargs)
        elapsed = time.perf_counter() - start
        logger.log(loglvl, f"{func.__qualname__} executed in {elapsed:.4f} seconds")
        return result

    return wrapper


@dataclasses.dataclass
class Range:
    """A range of addresses with a start (inclusive) and end (exclusive)."""

    start: int
    end: int

    def __post_init__(self):
        if self.start >= self.end:
            raise ValueError("start must be less than end")

    def __contains__(self, addr: int) -> bool:
        """Check if an address is within this range."""
        return self.start <= addr < self.end

    def __len__(self) -> int:
        """Return the size of the range in bytes."""
        return self.end - self.start

    def overlaps(self, other: "Range") -> bool:
        """Check if this range overlaps with another range."""
        return self.start < other.end and other.start < self.end

    def merge(self, other: "Range") -> "Range":
        return Range(min(self.start, other.start), max(self.end, other.end))


class IntervalSet:
    """
    Sorted, non-overlapping list of Range objects with O(log n) insertion.
    """

    __slots__ = ("_ranges",)

    def __init__(self) -> None:
        self._ranges: list[Range] = []

    def __iter__(self):
        return iter(self._ranges)

    def __len__(self):
        return len(self._ranges)

    # --- public ------------------------------------------------------------
    def add(self, new: Range) -> None:
        """
        Insert `new` and coalesce any overlaps / adjacencies in-place.
        """
        # Fast-path: first interval
        if not self._ranges:
            self._ranges.append(new)
            return

        # Binary-search insertion point by *start*
        idx = bisect_left(
            self._ranges, new.start, key=lambda r: r.start
        )  # Python 3.10+

        # Extend backward if necessary
        if idx > 0 and self._ranges[idx - 1].end >= new.start:
            idx -= 1

        # Merge forward while overlapping
        while idx < len(self._ranges) and new.overlaps(self._ranges[idx]):
            new = new.merge(self._ranges[idx])
            del self._ranges[idx]

        # Also coalesce "touching" intervals (…,end==new.start or vice-versa)
        if idx < len(self._ranges) and new.end == self._ranges[idx].start:
            new = new.merge(self._ranges[idx])
            del self._ranges[idx]
        if idx > 0 and self._ranges[idx - 1].end == new.start:
            new = new.merge(self._ranges[idx - 1])
            del self._ranges[idx - 1]
            idx -= 1

        self._ranges.insert(idx, new)

    # ­— optional helpers ---------------------------------------------------
    def covers(self, addr: int) -> bool:
        i = bisect_right(self._ranges, addr, key=lambda r: r.start) - 1
        return i >= 0 and addr < self._ranges[i].end

    def as_tuples(self):
        return [(r.start, r.end) for r in self._ranges]


def resolve_overlaps(ranges: list[Range]) -> IntervalSet:
    """
    Fast, linear-time overlap resolution: keep only the first chain
    whose start is ≥ the furthest end so far.
    """
    logger.info(f"Resolving overlaps among {len(ranges)} ranges")
    intervals = IntervalSet()

    for r in ranges:
        intervals.add(r)

        # decide whether to keep the chain object itself
        last_end = intervals.as_tuples()[-1][1]  # rightmost byte so far
        target = r.end
        if target == last_end:  # this chain extended the interval set
            logger.info(f"  Accepted (or widened): {r.start:X}-{r.end:X}")
        else:
            logger.info(f"  Rejected overlap: {r.start:X}-{r.end:X}")

    return intervals


class PatchManager:
    """Manages deferred patch operations."""

    class Mode(enum.Enum):
        PATCH = enum.auto()  # Use ida_bytes.patch_bytes
        PUT = enum.auto()  # Use ida_bytes.put_bytes

    def __init__(
        self,
        patch_mode: Mode = Mode.PATCH,
        dry_run: bool = False,
        auto_clear: bool = True,
    ):
        self.dry_run = dry_run
        self.patch_mode = patch_mode
        self.pending_patches: list[DeferredPatchOp] = []
        self.auto_clear = auto_clear
        logger.info(
            "PatchManager initialized (dry_run=%s, mode=%s)",
            self.dry_run,
            self.patch_mode.name,
        )

    def add_patch(self, address: int, byte_values: bytes):
        """Creates and queues a DeferredPatchOp."""
        op = DeferredPatchOp(address, byte_values, self.patch_mode)
        self.pending_patches.append(op)
        logger.debug("Queued patch operation: %s", op)

    def apply_all(self, dry_run_override: bool | None = None) -> bool:
        """Applies all queued patch operations."""
        logger.info("Applying %d queued patches...", len(self))
        success_count = 0
        fail_count = 0

        if dry_run_override is None:
            # None is a sentinel value here that represents "use the default"
            dry_run_override = self.dry_run

        for op in self.pending_patches:
            if op.apply(dry_run_override):
                success_count += 1
            else:
                fail_count += 1

        logger.info(
            "Patch application complete. Success: %d, Failed: %d",
            success_count,
            fail_count,
        )
        if self.auto_clear:
            self.pending_patches.clear()  # Clear the list after applying
        return fail_count == 0  # Return True if all patches were applied successfully

    def __len__(self) -> int:
        return len(self.pending_patches)


@dataclasses.dataclass(repr=False)
class DeferredPatchOp:
    """Class to store patch operations that will be applied later."""

    address: int
    byte_values: bytes
    mode: PatchManager.Mode
    dry_run: bool = False

    @classmethod
    def patch(cls, address: int, byte_values: bytes, dry_run: bool = False):
        return cls(address, byte_values, PatchManager.Mode.PATCH, dry_run)

    @classmethod
    def put(cls, address: int, byte_values: bytes, dry_run: bool = False):
        return cls(address, byte_values, PatchManager.Mode.PUT, dry_run)

    def apply(self, dry_run_override: bool = False) -> bool:
        """Apply the patch operation using either patch_bytes or put_bytes based on mode."""

        is_dry_run = dry_run_override or self.dry_run
        logger.info(
            "[*] %sPatching decrypted chunk %s at 0x%X (size: %d)",
            "(Dry Run) " if is_dry_run else "",
            ("revertably" if self.mode == PatchManager.Mode.PATCH else "destructively"),
            self.address,
            len(self.byte_values),
        )
        success = True
        if is_dry_run:
            return success

        try:
            import idaapi  # todo: make this a protocol + pass to the manager

            func = (
                idaapi.put_bytes
                if self.mode == PatchManager.Mode.PUT
                else idaapi.patch_bytes
            )
            func(self.address, self.byte_values)
        except Exception as e:
            logger.error(f"Failed to apply patch {self}: {e}", exc_info=True)
            success = False
        return success

    def __str__(self):
        """String representation with hex formatting."""
        dry_run_str = " (dry run)" if self.dry_run else ""
        return f"{self.__class__.__name__}({len(self.byte_values)} bytes, mode={self.mode.name}{dry_run_str} @ address=0x{self.address:X})"

    __repr__ = __str__


def make_chunks(buf_len: int, n_chunks: int, max_pat: int):
    """
    Yield exactly n_chunks tuples of
      (padded_start, padded_end, core_start, core_end).

    * core ranges partition [0, buf_len) evenly by floor division.
    * padded ranges extend each core by (max_pat-1) on both sides,
      clamped to [0, buf_len].
    """
    for i in range(n_chunks):
        # uniform core split
        core_start = (buf_len * i) // n_chunks
        core_end = (buf_len * (i + 1)) // n_chunks
        core_len = core_end - core_start

        # padding
        padded_start = max(0, core_start - (max_pat - 1))
        padded_end = min(buf_len, core_end + (max_pat - 1))
        padded_len = padded_end - padded_start

        logger.debug(
            "Chunk %2d/%d: "
            "core=[%#x-%#x) (%d bytes), "
            "padded=[%#x-%#x) (%d bytes)",
            i,
            n_chunks,
            core_start,
            core_end,
            core_len,
            padded_start,
            padded_end,
            padded_len,
        )

        yield padded_start, padded_end, core_start, core_end


@contextlib.contextmanager
def shm_buffer(
    name: str, buf_len: int | None = None
) -> typing.Generator[
    multiprocessing.shared_memory.SharedMemory | memoryview, None, None
]:
    """
    context manager to access the shared memory buffer.
    if buf_len is not provided, then the buffer will be the raw shm memory
    buffer else it will be a byte slice of the shm memory buffer.

    Usage:
        with shm_buffer(name=..) as buf:
            # use buf
    """
    shm = multiprocessing.shared_memory.SharedMemory(name=name)
    try:
        yield shm.buf[:buf_len] if buf_len else shm
    finally:
        shm.close()


def execute_chunk_with_shm_view(
    user_chunk_processor: typing.Callable[..., typing.Any],
    shm_name: str,
    padded_start: int,
    padded_end: int,
    *user_args: typing.Any,
) -> typing.Any:
    """
    Handles SHM attachment, memoryview creation, and cleanup for a user's chunk processing function.

    Ensures that the memoryview of the chunk is deleted before the shared memory object is closed,
    preventing "BufferError: cannot close exported pointers exist".

    Args:
        user_chunk_processor: A callable that takes (memoryview, *user_args) and returns results.
                              The memoryview passed to it is a slice of the shared memory.
        shm_name: Name of the shared memory segment.
        padded_start: Start offset for the memoryview slice within the shared memory buffer.
        padded_end: End offset for the memoryview slice.
        *user_args: Additional arguments to pass to the user_chunk_processor.

    Returns:
        The result from the user_chunk_processor.
    """
    # shm_buffer is called with no buf_len, so it yields the SharedMemory object itself.
    with shm_buffer(shm_name) as shm_object:  # shm_object is a SharedMemory instance
        if not isinstance(shm_object, multiprocessing.shared_memory.SharedMemory):
            # This case should ideally not happen if shm_buffer is used as intended (no buf_len)
            # or if shm_buffer is modified to always yield SharedMemory object and handle buf internally.
            # For now, assume shm_object.buf is the way if shm_buffer yields memoryview directly.
            # However, the current shm_buffer yields SharedMemory if buf_len is None.
            logger.error(
                "shm_buffer did not yield a SharedMemory object as expected. Type: %s",
                type(shm_object),
            )
            # Fallback or raise error, this indicates misuse or change in shm_buffer's contract
            # For robustness, let's try to access .buf if it's not SharedMemory but has it
            try:
                buffer_to_view = shm_object.buf  # type: ignore
            except AttributeError:
                logger.error("Shared memory object does not have a .buf attribute.")
                raise TypeError(
                    "shm_buffer yielded an unexpected object type without a .buf attribute."
                ) from None
        else:
            buffer_to_view = shm_object.buf

        # Create the specific memoryview for this chunk
        # The memoryview is taken from the *entire* buffer of the shm_object
        chunk_mv = memoryview(buffer_to_view)[padded_start:padded_end]
        try:
            # Call the user's actual logic function
            result = user_chunk_processor(chunk_mv, *user_args)
            return result
        finally:
            # Crucial: delete the memoryview to release buffer export before shm_object.close()
            # which is called by shm_buffer's __exit__.
            del chunk_mv
            logger.debug(
                "Deleted memoryview for chunk [%d:%d] from shm '%s'",
                padded_start,
                padded_end,
                shm_name,
            )
